

/*
#    Sfront, a SAOL to C translator    
#    This file: Code generation: declarations/initializations
#    Copyright (C) 1999  Regents of the University of California
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License (Version 2) as
#    published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#    Maintainer: John Lazzaro, lazzaro@cs.berkeley.edu
*/

#include "tree.h"
#include "parser.tab.h"


void makeargcargv(void);
void makepreamble(void);   
void makeargcargv(void);   
void makesysinfo(void);    
void corefunctions(void);  
void makebusnames(void);   
void makeglobalvars(void); 
void makeglobaltabs(void); 
void instrvars(int *); 
void instrtabs(int *); 
void effvars(int *);   
void efftabs(int *);    
void makefexterns(void);   
void printinstrdata(void); 
void make_ninstr(int *, int *); 
void audiolibrary(void);   
void controllibrary(void); 


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                       void preamble()                        */
/*                                                              */
/* This function prints out the start of the sa.c file --       */
/* everything up to the code generation for opcodes. Some       */
/* of the functions it calls are defined as externs above.      */
/*                                                              */
/*______________________________________________________________*/


void preamble(void)

{
  int mvar = 1;     /* sets size of instr variable array       */
  int mtab = 1;     /* sets size of instr table array          */


  makepreamble();   /* generates lib/csrc/preamble.c constants */
  makeargcargv();   /* argc/argc info for sa.c and sfront      */

  makesysinfo();    /* all constants and data structures for   */
                    /* sound engine that are known early on    */

  corefunctions();  /* prints out fixed functions used for     */
                    /* certain core opcodes/generators.        */

  makebusnames();   /* prints out audio bus structures         */

  makeglobalvars(); /* prints out global variable structures   */
                    /* including many of the MIDI variables    */
  makeglobaltabs(); /* prints out the global tables            */

  instrvars(&mvar); /* symbol variables for non-effects instrs */
  instrtabs(&mtab); /* symbol tables for effects instrs        */
  effvars(&mvar);   /* symbol variables for non-effects instrs */
  efftabs(&mtab);   /* symbol tables for effects instrs        */

  makefexterns();   /* all extern function headers             */

  printinstrdata(); /* print note-list data and structures     */
                    /* include driver reflection mechanism     */

  make_ninstr(&mvar,&mtab); /* create the instr variable stack */

  audiolibrary();   /* include -aout and -ain libraries        */
  controllibrary(); /* include -cin libraries                  */

}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                                                              */
/* The second-level functions called by preamble. Some helper   */
/* functions are defined about the function itself, others      */
/* defined in a utility function section that ends the file.    */
/*                                                              */
/*______________________________________________________________*/


/****************************************************************/
/*            prints out argc/argv for sfront and sa.c          */
/****************************************************************/

void makeargcargv(void) 

{

  int i;

  fprintf(outfile,"int asys_argc;\n");
  fprintf(outfile,"char ** asys_argv;\n\n");
  
  fprintf(outfile,"int csys_argc;\n");
  fprintf(outfile,"char ** csys_argv;\n\n");
  
  fprintf(outfile,"\nint csys_sfront_argc = %i;\n", sfront_argc);
  fprintf(outfile,"\nchar * csys_sfront_argv[%i] = {\n", sfront_argc);

  i = 0;
  while (i < sfront_argc)
    {
      fprintf(outfile,"\"%s\"", sfront_argv[i++]);
      if (i < sfront_argc)
	fprintf(outfile,",\n");
      else
	fprintf(outfile,"};\n\n\n");
    }

}

/****************************************************************/
/*  constants and datastructures of sound engine (known early)  */
/****************************************************************/

void makesysinfo(void) 

{
  sigsym * sptr;
  float minlatency, maxlatency, defaultlatency;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* application ID, network status  */
  /*_________________________________*/

  fprintf(outfile,"#define APPNAME \"sfront\"\n");
  fprintf(outfile,"#define APPVERSION \"%s\"\n", IDSTRING);
  fprintf(outfile,"#define NSYS_NET %i\n", session ? 1 : 0);

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* all the temporal constants    */
  /*_______________________________*/

  fprintf(outfile,"#define ARATE %i.0F\n", srate);
  fprintf(outfile,"#define ATIME %eF\n", 1.0F/srate);
  fprintf(outfile,"#define KRATE %i.0F\n", krate);
  fprintf(outfile,"#define KTIME %eF\n", 1.0F/krate);
  fprintf(outfile,"#define KMTIME %eF\n", 1000.0F/krate);
  fprintf(outfile,"#define KUTIME %iL\n", (int)((1000000.0F/krate)+0.5F));
  fprintf(outfile,"#define ACYCLE %iL\n", srate/krate);

  fprintf(outfile,"float tempo = %i.0F;\n", confmidi->miditicks ? 120 : 60);
  fprintf(outfile,"float scoremult = %eF;\n\n",1.0F/krate);

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* symbols for audio I/O         */
  /*_______________________________*/

  fprintf(outfile,"#define ASYS_RENDER   0\n");
  fprintf(outfile,"#define ASYS_PLAYBACK 1\n");
  fprintf(outfile,"#define ASYS_TIMESYNC 2\n\n");
  fprintf(outfile,"#define ASYS_SHORT   0\n");
  fprintf(outfile,"#define ASYS_FLOAT   1\n\n");

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* the audio output variables    */
  /*_______________________________*/

  fprintf(outfile,"/* audio i/o */\n\n");
  fprintf(outfile,"#define ASYS_OCHAN %iL\n",outputbus->width);

  switch (makeaudiotypeout(aout)) {
  case SAMPLE_SHORT:
    fprintf(outfile,"#define ASYS_OTYPENAME  ASYS_SHORT\n");
    fprintf(outfile,"#define ASYS_OTYPE  short\n");
    break;
  case SAMPLE_FLOAT:
    fprintf(outfile,"#define ASYS_OTYPENAME  ASYS_FLOAT\n");
    fprintf(outfile,"#define ASYS_OTYPE  float\n");
    break;
  }

  if ((ainflow == PASSIVE_FLOW) && (aoutflow == PASSIVE_FLOW))
    {
      fprintf(outfile,"long asys_osize = 0;\n");
      fprintf(outfile,"long obusidx = 0;\n\n");
      fprintf(outfile,"ASYS_OTYPE * asys_obuf = NULL;\n\n");
    }
  else
    {
      fprintf(outfile,"long asys_exit_status = ASYS_DONE;\n\n");
    }

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* the audio input variables     */
  /*_______________________________*/

  sptr = getvsym(&busnametable,"input_bus");
  if ((sptr != NULL) && (inchannels > 0))
    {
      fprintf(outfile,"#define ASYS_ICHAN %iL\n",inchannels);

      switch (makeaudiotypein(ain)) {
      case SAMPLE_SHORT:
	fprintf(outfile,"#define ASYS_ITYPENAME  ASYS_SHORT\n");
	fprintf(outfile,"#define ASYS_ITYPE  short\n");
	break;
      case SAMPLE_FLOAT:
	fprintf(outfile,"#define ASYS_ITYPENAME  ASYS_FLOAT\n");
	fprintf(outfile,"#define ASYS_ITYPE  float\n");
	break;
      }

      if ((ainflow == PASSIVE_FLOW) && (aoutflow == PASSIVE_FLOW))
	{
	  fprintf(outfile,"long asys_isize = 0;\n");
	  fprintf(outfile,"long ibusidx = 0;\n\n");
	  fprintf(outfile,"ASYS_ITYPE * asys_ibuf = NULL;\n\n");
	}
    }

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /*   active audio driver setup   */
  /*_______________________________*/

  if ((ainflow == ACTIVE_FLOW) || (aoutflow == ACTIVE_FLOW))
    {
      if ((sptr != NULL) && (inchannels > 0))
	{
	  fprintf(outfile, "#define ASYS_ACTIVE_IO\n\n");
	  fprintf(outfile, "extern int asys_iorun(ASYS_ITYPE ibuf[],"
		  " long * isize,\n");
	  fprintf(outfile, "	                  ASYS_OTYPE obuf[],"
		  " long * osize);\n\n");
	}
      else
	{
	  fprintf(outfile, "#define ASYS_ACTIVE_O\n\n"); 
	  fprintf(outfile, "extern int asys_orun(ASYS_OTYPE obuf[],"
		  " long * osize);\n\n");
	}
    }

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /*   latency and timeoption      */
  /*_______________________________*/

  /* print out latency as number of samples */
  /* will always be a power of 2 */

  fprintf(outfile,"#define ASYS_USERLATENCY  %i\n", (latency != -1));
  fprintf(outfile,"#define ASYS_LOWLATENCY   0\n");
  fprintf(outfile,"#define ASYS_HIGHLATENCY  1\n");

  if ((timeoptions == TIMESYNC) || (clatency == LOW_LATENCY_DRIVER) ||
      (ainlatency == LOW_LATENCY_DRIVER) || session)
    {
      fprintf(outfile,"#define ASYS_LATENCYTYPE  ASYS_LOWLATENCY\n");
      minlatency = LOW_LATENCY_MIN;
      defaultlatency = LOW_LATENCY_DEFAULT;
      maxlatency = LOW_LATENCY_MAX;
    }
  else
    {
      fprintf(outfile,"#define ASYS_LATENCYTYPE  ASYS_HIGHLATENCY\n");
      minlatency = HIGH_LATENCY_MIN;
      defaultlatency = HIGH_LATENCY_DEFAULT;
      maxlatency = HIGH_LATENCY_MAX;
    }

  if (latency == -1)
    latency = defaultlatency;
  else
    {
      if (latency < minlatency)
	{
	  latency = minlatency;
	  printf("Warning: Requested latency too small, using (%fs).\n\n",
		 latency);
	}
      if (latency > maxlatency)
	{	    
	  printf("Warning: Using requested latency (%fs), but\n",
		 latency);
	  printf("its probably too large and will cause artifacts.\n\n");
	}
    }

  fprintf(outfile,"#define ASYS_LATENCY %fF\n", latency);  

  switch(timeoptions) {
  case RENDER:
    fprintf(outfile,"#define ASYS_TIMEOPTION ASYS_RENDER\n\n");
    break;
  case PLAYBACK:
    fprintf(outfile,"#define ASYS_TIMEOPTION ASYS_PLAYBACK\n\n");
    break;
  case TIMESYNC:
    fprintf(outfile,"#define ASYS_TIMEOPTION ASYS_TIMESYNC\n\n");
    break;
  }

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* the struct instr_line         */
  /*_______________________________*/


  fprintf(outfile,"#define MAXPFIELDS %i\n\n", numpfields);

  fprintf(outfile,"struct instr_line { \n\n");
  fprintf(outfile,"float starttime;  /* score start time of note */\n");
  fprintf(outfile,"float endtime;    /* score end time of note */\n");
  fprintf(outfile,"float startabs;   /* absolute start time of note */\n");
  fprintf(outfile,"float endabs;     /* absolute end time of note */\n");
  fprintf(outfile,"float abstime;    /* absolute time extension */\n");
  fprintf(outfile,"float time;       /* time of note start (absolute) */\n");
  fprintf(outfile,"float itime;      /* elapsed time (absolute) */\n");
  fprintf(outfile,"float sdur;       /* duration of note in score time*/\n\n");
  fprintf(outfile,"int kbirth;       /* kcycleidx for note launch */\n");
  fprintf(outfile,"int released;     /* flag for turnoff*/\n");
  fprintf(outfile,"int turnoff;      /* flag for turnoff */\n");
  fprintf(outfile,"int noteon;       /* NOTYETUSED,TOBEPLAYED,PAUSED,PLAYING,ALLDONE */\n");
  fprintf(outfile,"int notestate;    /* index into state array */\n");
  fprintf(outfile,"int launch;       /* only for dynamic instruments */\n");
  fprintf(outfile,"int numchan;      /* only for MIDI notes */\n");
  fprintf(outfile,"int preset;       /* only for MIDI notes */\n");
  fprintf(outfile,"int notenum;      /* only for MIDI notes */\n");
  fprintf(outfile,"int label;        /* SASL label index + 1, or 0 (no label) */\n\n");
  fprintf(outfile,"                  /* for static MIDI, all-sounds noteoff */\n\n");
  fprintf(outfile,"float p[MAXPFIELDS];          /* parameters */\n\n");
  fprintf(outfile,"struct ninstr_types * nstate; /* pointer into state array */\n\n");

  fprintf(outfile,"};\n\n");


  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* maximum depth of any oparray chain         */
  /*____________________________________________*/

  if (maxoparraydepth)
    fprintf(outfile,"#define MAXOPARRAY %i\n", maxoparraydepth);
    
}

/****************************************************************/
/*    hand-coded functions for core opcode/gens . Should be a   */
/*    library eventually.                                       */
/****************************************************************/

void corefunctions(void)

{
  int i, tsize;

  if (has.w_window)
    {
      fprintf(outfile,"\n/* for window wavetable generator */\n\n");
      fprintf(outfile,"double modbessel(double x)\n\n");
      fprintf(outfile,"{\n");
      fprintf(outfile,"  double sum,a;\n\n");
      fprintf(outfile,"  sum = a = 1.0;\n");
      fprintf(outfile,"  a *= x*x; sum += a*2.5e-1;\n");
      fprintf(outfile,"  a *= x*x; sum += a*1.5625e-2;\n");
      fprintf(outfile,"  a *= x*x; sum += a*4.340278e-4;\n");
      fprintf(outfile,"  a *= x*x; sum += a*6.781684e-6;\n");
      fprintf(outfile,"  a *= x*x; sum += a*6.781684e-8; \n");
      fprintf(outfile,"  a *= x*x; sum += a*4.709503e-10;\n");
      fprintf(outfile,"  a *= x*x; sum += a*2.402808e-12; \n");
      fprintf(outfile,"  a *= x*x; sum += a*9.385967e-15;\n");
      fprintf(outfile,"  a *= x*x; sum += a*2.896903e-17;\n");
      fprintf(outfile,"  a *= x*x; sum += a*7.242258e-20;\n\n");
      fprintf(outfile,"  return sum;\n");
      fprintf(outfile,"}\n\n");
    }

  if (has.o_grain)
    {
      fprintf(outfile,"\n/* for grain core opcode */\n\n");
      fprintf(outfile,"#define GRLOSCIL 0\n");
      fprintf(outfile,"#define GRDOSCIL 1\n");
      fprintf(outfile,"#define GROSCIL  2\n");
      fprintf(outfile,"#define GRTIME   0\n");
      fprintf(outfile,"#define GRPHASE  1\n");
      fprintf(outfile,"#define GRFREQ   2\n");
      fprintf(outfile,"#define GRAMP    3\n");
      fprintf(outfile,"#define GRDUR    4\n");
      fprintf(outfile,"#define GRINVDUR 5\n");
      fprintf(outfile,"#define GRSTATE  6\n");
      fprintf(outfile,"#define GRNUM   256\n\n");
    }

  if (has.o_frac)
    {
      fprintf(outfile,"double fracdummy;\n\n");
    }

  if (has.o_flange)
    {
      fprintf(outfile,"#define FLNOFF 0.002F\n\n");
    }

  if (has.o_chorus)
    {
      fprintf(outfile,"#define CHROFF 0.020F\n\n");
    }

  if (has.w_sample)
    {      
      fprintf(outfile,"FILE * sfile = NULL;\n\n");
      /* add rfseek later ... */
    }

  if (has.o_speedt)
    makerunspt();

  if (has.o_buzz)
    {
      tsize = 2048;
      fprintf(outfile,"\n#define TRIGSIZE %i\n",tsize);
      fprintf(outfile,"\n#define TRIGQUAD %i\n",tsize/4);
      fprintf(outfile,"\n#define TRIGSIZEF %i.0F\n",tsize);
      fprintf(outfile,"float sintab[TRIGSIZE+TRIGQUAD] = { \n");
      for (i=0;i<tsize;i++)
	{
	  fprintf(outfile,"%10eF", sin((6.283185/tsize)*i));
	  fprintf(outfile,", ");
	  if ((i % 8) == 0)
	    fprintf(outfile,"\n");
	}
      for (i=0;i<(tsize/4);i++)
	{
	  fprintf(outfile,"%10eF", sin((6.283185/tsize)*i));
	  if (i != ((tsize/4)-1))
	    fprintf(outfile,", ");
	  if ((i % 8) == 0)
	    fprintf(outfile,"\n");
	}
      fprintf(outfile,"};\n\n");
      fprintf(outfile,"#define TSIN(x) sintab[(int)(x)]\n");
      fprintf(outfile,"#define TCOS(x) sintab[(int)(x)+TRIGQUAD]\n");
      fprintf(outfile,"#define BUZZDIVISOR 0.02F\n");
      fprintf(outfile,"#define BUZZMINVAL 0.995F\n");
      fprintf(outfile,"#define BUZZMAXVAL 1.005F\n");
    }

  if (has.o_fft || has.o_ifft)
    {
      tsize = 4;
      while (tsize <= 8192)
	{
	  fprintf(outfile,"extern float fft%itab[]; \n",tsize);
	  fprintf(outfile,"extern int fft%imap[]; \n", tsize);
	  tsize *= 2;
	}
      fprintf(outfile,"\n");
    }
}

/****************************************************************/
/*    this isn't called in preamble, but in writeorc. since     */
/*    the fft tables are huge, we only print those we need      */
/****************************************************************/

void postcorefunctions(void)

{
  int i, tsize, ds, dc, j, tab;

  if (has.o_fft || has.o_ifft)
    {
      tsize = 4;
      tab = 2;
      while (tsize <= 8192)
	{
	  if (ffttables[tab])
	    {
	      fprintf(outfile,"float fft%itab[%i] = { \n",tsize,3*(tsize>>2));
	      for (i=0;i<3*(tsize>>2);i++)
		{
		  fprintf(outfile,"%10eF", cos((6.283185/tsize)*i));
		  if (i != (3*(tsize>>2)-1))
		    fprintf(outfile,", ");
		  if ((i % 8) == 0)
		    fprintf(outfile,"\n");
		}
	      fprintf(outfile,"};\n\n");
	      
	      fprintf(outfile,"int fft%imap[%i] = { \n", tsize, tsize);
	      for (i=0;i< tsize;i++)
		{
		  j=0;
		  ds = tsize >> 1;
		  dc = 1;
		  while (ds)
		    {
		      j |= (ds&i) ? dc : 0;
		      ds = ds >> 1;
		      dc = dc << 1;
		    }	      
		  fprintf(outfile,"%i", j);
		  if (i != tsize - 1)
		    fprintf(outfile,", ");
		  if ((i % 8) == 0)
		    fprintf(outfile,"\n");
		}
	      fprintf(outfile,"};\n\n");
	    }

	  tab++;
	  tsize *= 2;
	}
    }
}

/****************************************************************/
/*            prints out the audio bus structures               */
/****************************************************************/

void makebusnames(void) 

{
  sigsym * sptr;
  int i;

  sptr = busnametable;
  i = 0;
  while (sptr != NULL)
    {
      fprintf(outfile,"#define BUS_%s %i\n",sptr->val,i);
      i = i + sptr->width;
      if ((sptr == outputbus) || !strcmp(sptr->val, "input_bus"))
	fprintf(outfile,"#define ENDBUS_%s %i\n",sptr->val,i);
      sptr = sptr->next;
    }
  fprintf(outfile,"#define ENDBUS %i\n",i);
  fprintf(outfile,"float bus[ENDBUS];\n");
  
  if (useshadowbus)
    fprintf(outfile,"float sbus[ENDBUS];\n");   /* shadow bus */

  fprintf(outfile,"\n");
}

/*****************************************************************/
/* this function is defined in the utility section; its prints   */
/*    the symbol defines of opcode variables and decendents.     */
/*****************************************************************/

extern void opcodedefines(tnode * optr, char * prefix, 
			  int * i, int * j);


/****************************************************************/
/*      prints out the global variables + some MIDI variables   */
/****************************************************************/

void makeglobalvars(void) 

{  
  int i, j;
  sigsym * sptr;
  sigsym * bptr;

  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /*  we first print out consts    */
  /*  for every global signal      */
  /*  variable, leaving space      */
  /*  at the top for MIDI vars.    */
  /*_______________________________*/


  /* calculate MIDI space, and define midimasterchannel space */

  totmidichan = confmidi->midinumchan + sstrmidi->midinumchan;
  if (cmidi || session)
    {
      fprintf(outfile,"#define CSYS_EXTCHANSTART %i\n",totmidichan);
      fprintf(outfile,"int midimasterchannel = CSYS_EXTCHANSTART;\n\n");
      totmidichan += cmaxchan;
    }
  else
    {
      if (totmidichan)
	fprintf(outfile,"int midimasterchannel = 0;\n\n");
      else
	fprintf(outfile,"float fakeMIDIctrl[256];\n\n");
    }

  i = totmidichan*MIDIFRAMELEN;

  if (i>0)
    fprintf(outfile,"/* first %i locations for MIDI */\n",i);
  fprintf(outfile,"#define GBL_STARTVAR %i\n",i);


  /* actually print out global signal variable consts */

  sptr = globalsymtable;
  while (sptr != NULL)
    {
      fprintf(outfile,"#define GBL_%s %i\n",sptr->val,i);

      /* tables done during TBL_GBL */

      if (csasl && (sptr->vartype!=TABLETYPE))
	{
	  if (bitfile && bitsymin)
	    {
	      if ((bptr = getvsym(&bitsymin,sptr->val)))
		fprintf(outfile,"#define CSYS_SASL_GBL_%s %i\n",
			bptr->defnode->val, i);
	    }
	  else
	    fprintf(outfile,"#define CSYS_SASL_GBL_%s %i\n",sptr->val,i);
	}

      i = i + sptr->width;
      sptr = sptr->next;
    }

  /* print out signal variable consts for opcodes called in tables */

  j = 1;
  currinstrument = NULL;
  opcodedefines(globalopcodecalls,"GBL",&i,&j);
  fprintf(outfile,"#define GBL_ENDVAR %i\n",i);


  /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
  /* We then print out the         */
  /* initialized constant          */
  /* space for globals. Note       */
  /* only MIDI has non-zero        */
  /* values now, but startup       */
  /* could be executed in sfront   */
  /* to provide additional inits.  */
  /*_______________________________*/

  fprintf(outfile,"/* global variables */\n\n");

  fprintf(outfile, "dstack global[GBL_ENDVAR+1];\n\n");

  if (maxoparraydepth)
    {  
      fprintf(outfile,"dstack * globalvstate[MAXOPARRAY];\n");
      fprintf(outfile,"struct tableinfo * globaltstate[MAXOPARRAY];\n");
    }

}

/*****************************************************************/
/* this function is defined in the utilities section; it prints  */
/* the symbol defines for tables of an opcode and its decendents.*/
/*****************************************************************/

void opcodetabledefines(tnode * optr, char * prefix, 
			int * i, int * j);


/****************************************************************/
/*              prints out the global tables                    */
/****************************************************************/

void makeglobaltabs(void) 

{  
  int i, j;
  sigsym * sptr;
  sigsym * bptr;

  sptr = globalsymtable;
  i = 0;
  while (sptr != NULL)
    {
      if (sptr->vartype == TABLETYPE)
	{
	  fprintf(outfile,"#define TBL_GBL_%s %i\n",sptr->val,i);
	  if (csasl)
	    {
	      if (bitfile && bitsymin)
		{
		  bptr = getvsym(&bitsymin,sptr->val);

		  /* bptr is null for synthetic global tables */
		  
		  if (bptr)
		    fprintf(outfile,"#define CSYS_SASL_GBL_%s %i\n",
			    bptr->defnode->val, i);
		}
	      else
		if ((sptr->val[0]!='_'))
		  fprintf(outfile,"#define CSYS_SASL_GBL_%s %i\n",
			  sptr->val,i);
	    }
	  i++;  
	}
      sptr = sptr->next;
    }
  j = 1;
  opcodetabledefines(globalopcodecalls,"GBL",&i,&j);
  fprintf(outfile,"#define GBL_ENDTBL %i\n",i);
  fprintf(outfile,"struct tableinfo gtables[GBL_ENDTBL+1];\n\n");


}


/*****************************************************************/
/*       this function is defined in the utility section.        */
/*       it prints out extra symbols for the csys driver         */
/*****************************************************************/

extern void instrdriversymbols(sigsym *, sigsym *, int i);


/*****************************************************************/
/*       this function is defined in the utility section.        */
/*       it determines the true width of a symbol entry          */
/*****************************************************************/

extern int symbolwidth(int);


/****************************************************************/
/*   prints out variable symbols for non-effects instrs         */
/****************************************************************/

void instrvars(int * maxvars) 

{  
  int i, j, k;
  sigsym * sptr;
  sigsym * iptr;

  j = 1;
  iptr = instrnametable;
  while (iptr != NULL)
    {
      if ((iptr->score)||(iptr->ascore)||(iptr->dyn)||(iptr->startup)
	  ||(iptr->midi)||(iptr->amidi)
	  ||((cmidi || session) && iptr->miditag)|| csasl)
	{
	  i = 0;
	  currinputwidth = 1;
	  currinstrwidth = iptr->width;
	  currinstrument = iptr;
	  sptr = iptr->defnode->sptr;
	  while (sptr != NULL)
	    {
	      if ((sptr->vartype != TMAPTYPE) &&
		  (sptr->tref->mirror == REQUIRED))
 		{
		  fprintf(outfile,"#define %s_%s %i\n",iptr->val,sptr->val,i);
		  if (csasl)
		    instrdriversymbols(sptr, iptr, i);
		  i += symbolwidth(sptr->width);
		}
	      sptr = sptr->next;
	    }
	  for (k = 0; k < iptr->cref->ifstate; k++)
	    fprintf(outfile,"#define %s__sym_if%i %i\n", iptr->val, k, i++);
	  /* currinputwidth = 0; */   /* don't know why this was here ... */
	  opcodedefines(iptr->defnode->optr,iptr->val,&i,&j);
	  j = 1;
	  fprintf(outfile,"#define %s_ENDVAR %i\n\n",iptr->val,i);
	  if (i > (*maxvars))
	    *maxvars = i;
	}
      iptr = iptr->next;
    }
  currinstrument = NULL;

}


/****************************************************************/
/*   prints out table symbols for non-effects instrs         */
/****************************************************************/

void instrtabs(int * maxtables) 

{    
  int i, j;
  sigsym * sptr;
  sigsym * iptr;

  j = 1;
  iptr = instrnametable;
  while (iptr != NULL)
    {
      if ((iptr->score)||(iptr->ascore)||(iptr->dyn)||(iptr->startup)
	  ||(iptr->midi)||(iptr->amidi)
	  ||((cmidi || session) && iptr->miditag)|| csasl)
	{
	  sptr = iptr->defnode->sptr;
	  i = 0;
	  while (sptr != NULL)
	    {
	      if (sptr->vartype == TABLETYPE)
		{
		  fprintf(outfile,"#define TBL_%s_%s %i\n",
			  iptr->val,sptr->val,i);
		  i++;
		}
	      sptr = sptr->next;
	    }
	  opcodetabledefines(iptr->defnode->optr,iptr->val,&i,&j);
	  j = 1;
	  fprintf(outfile,"#define %s_ENDTBL %i\n\n",iptr->val,i);
	  if (i > (*maxtables))
	    *maxtables = i;
	}
      iptr = iptr->next;
    }

}

/****************************************************************/
/*   prints out variable symbols for effects instrs         */
/****************************************************************/

void effvars(int * maxvars) 

{

  sigsym * sptr;
  tnode * tptr;
  int i, j, k;

  j = 1;
  tptr = instances;
  while (tptr != NULL)
    {
      i = 0;
      currinputwidth = tptr->inwidth;
      currinstrwidth = tptr->sptr->width;
      currinstrument = tptr->sptr;
      sptr = tptr->sptr->defnode->sptr;
      while (sptr != NULL)
	{	    	      
	  if ((sptr->vartype != TMAPTYPE) &&
	      (sptr->tref->mirror == REQUIRED)) 
	    {
	      fprintf(outfile,"#define %s_%s %i\n",tptr->val,sptr->val,i);
	      i += symbolwidth(sptr->width);
	    }
	  sptr = sptr->next;
	}
      for (k = 0; k < tptr->sptr->cref->ifstate; k++)
	fprintf(outfile,"#define %s__sym_if%i %i\n", tptr->val, k, i++);
      opcodedefines(tptr->sptr->defnode->optr,tptr->val,&i,&j);
      j = 1;
      fprintf(outfile,"#define %s_ENDVAR %i\n\n",tptr->val,i);
      if (i > (*maxvars))
	*maxvars = i;
      tptr = tptr->next;
    }
  currinstrument = NULL;

}


/****************************************************************/
/*   prints out table symbols for effects instrs         */
/****************************************************************/

void efftabs(int * maxtables) 

{    
  tnode * tptr;
  sigsym * sptr;
  int i, j;

  j = 1;
  tptr = instances;
  while (tptr != NULL)
    {
      sptr = tptr->sptr->defnode->sptr;
      i = 0;
      while (sptr != NULL)
	{
	  if (sptr->vartype == TABLETYPE)
	    {
	      fprintf(outfile,"#define TBL_%s_%s %i\n",tptr->val,sptr->val,i);
	      i++;
	    }
	  sptr = sptr->next;
	}
      opcodetabledefines(tptr->sptr->defnode->optr,tptr->val,&i,&j);
      j = 1;
      fprintf(outfile,"#define %s_ENDTBL %i\n\n",tptr->val,i);
      if (i > (*maxtables))
	*maxtables = i;
      tptr = tptr->next;
    }

}

/*****************************************************************/
/* these functions are defined in the utility section at the     */
/* bottom of this file. they print out function headers so that  */
/*              correct compilation can occur.                   */
/*****************************************************************/

extern void printopexterns(tnode * tptr);
extern void printpassexterns(void);  
extern void printfinputexterns(void);
extern void printwavetableexterns(void);

/****************************************************************/
/*        prints out all extern function headers                */
/****************************************************************/

void makefexterns(void) 

{  
  sigsym * sptr;
  tnode * tptr;

  /* signal handlers */

  if (catchsignals)
    {
      fprintf(outfile,"extern void signal_handler(int);\n");
    }
  fprintf(outfile,"extern void sigint_handler(int);\n\n");

  /* headers for an instrs */

  sptr = instrnametable;
  while (sptr != NULL)
    {
      if ((sptr->score)||(sptr->ascore)||(sptr->dyn)||(sptr->startup)
	  ||(sptr->midi)||(sptr->amidi)
	  ||((cmidi || session) && sptr->miditag)||csasl)
	{
	  currinstance = NULL;
	  currinstancename = sptr->val;
	  printopexterns(sptr->defnode->optr);
	}
      sptr = sptr->next;
    }

  /* and effects instrs */

  tptr = instances;
  while (tptr != NULL)
    {
      currinstance = tptr;
      currinstancename = tptr->val;
      printopexterns(tptr->sptr->defnode->optr);
      tptr = tptr->next;
    }

  /* and opcodes used in the global block */
			
  currinstancename = "GBL";
  currinstance = NULL;
  printopexterns(globalopcodecalls);

  printpassexterns();           /* and the _{i,k,a}pass functions */
  printfinputexterns();        /* and audio input functions       */
  printsaoltables(S_EXPORTS);  /* and constant wavetable arrays   */

}

/*****************************************************************/
/*    these functions are defined in the utility section.        */
/*         they print out note data and structures               */
/* for the instrument types, and the reflection data structures  */
/*****************************************************************/

void printcdriverpreamble(void);
void printinstrlists(int *, int *, int *);
void printcdriverlists(int *);
void printdynamiclists(void);


/****************************************************************/
/*                 print instr declarations                     */
/****************************************************************/

void printinstrdata(void)

{

  int totlines = 1;
  int usesdyn = 0;
  int instrnum = 0;

  initendtime();                     /*    initializes endtime variable   */
  printcdriverpreamble();            /* print control driver support vars */

  printinstrlists(&totlines, &usesdyn, &instrnum);   /* makes instr lists */

  if (cin || session)
    printcdriverlists(&instrnum); /* makes control driver MIDI/SASL lists */

  initscorecontrol(RELTSTAMP);         /* control, tempo, and table lists */
  initscorecontrol(ABSTSTAMP);
  initscoretempo(RELTSTAMP);
  initscoretempo(ABSTSTAMP);
  initscoretable(RELTSTAMP);
  initscoretable(ABSTSTAMP);

  /* heuristic for setting amount of note/line space */

  if (usesdyn || cin || session)
    {
      if (MAXDNOTES > 3*totlines)
	{
	  fprintf(outfile,"#define MAXLINES %i\n",MAXDNOTES);
	  fprintf(outfile,"#define MAXSTATE %i\n",MAXDNOTES + 4*totlines);
	}
      else
	{
	  fprintf(outfile,"#define MAXLINES %i\n",4*totlines);
	  fprintf(outfile,"#define MAXSTATE %i\n",4*totlines);
	}
    }
  else
    fprintf(outfile,"#define MAXSTATE %i\n",totlines);

  fprintf(outfile,"\n");

  if (usesdyn)
    printdynamiclists();  /* makes dynamic instrument lists */


}

/****************************************************************/
/*            create the instr variable stack                   */
/****************************************************************/

void make_ninstr(int * maxvars, int * maxtables)

{

  fprintf(outfile,"\n");
  fprintf(outfile,"#define MAXVARSTATE %i\n",*maxvars);
  fprintf(outfile,"#define MAXTABLESTATE %i\n",*maxtables);
  fprintf(outfile,"\n");

  fprintf(outfile,"/* ninstr: used for score, effects, */\n");
  fprintf(outfile,"/* and dynamic instruments          */\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"struct ninstr_types {\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"struct instr_line * iline; /* pointer to score line */\n");
  fprintf(outfile,"dstack v[MAXVARSTATE];     /* parameters & variables*/\n");
  fprintf(outfile,"struct tableinfo t[MAXTABLESTATE]; /* tables */\n");

  if (maxoparraydepth)
    {  
      fprintf(outfile,"dstack * vstate[MAXOPARRAY];\n");
      fprintf(outfile,"struct tableinfo * tstate[MAXOPARRAY];\n");
    }

  fprintf(outfile,"\n");
  fprintf(outfile,"} ninstr[MAXSTATE];\n");
  fprintf(outfile,"\n");

}


/*****************************************************************/
/*       these function is defined in the utility section.       */
/*       they print out the audio input and output libraries     */
/*****************************************************************/

extern void makeoutputaudiodriver(sigsym * inputbus);
extern void makeinputaudiodriver(sigsym * inputbus);

/****************************************************************/
/*             loads in appropriate audio library               */
/****************************************************************/

void audiolibrary(void)

{
  sigsym * inputbus;
  int i, j;

  /********************************/
  /* print CDRIVER, ADRIVER names */
  /********************************/

  if (cin)
    {
      fprintf(outfile,"#define CSYS_CDRIVER_");
      for (i=0; i< (int)strlen(cinname); i++)
	fprintf(outfile, "%c", toupper(cinname[i]));
      fprintf(outfile,"\n");
    }

  fprintf(outfile,"#define ASYS_OUTDRIVER_");
  j = 0;
  for (i=0; i< (int)strlen(aoutname); i++)
    if (aoutname[i] == '.')
      {
	j = i + 1;
	break;
      }
  for (i=j; i< (int)strlen(aoutname); i++)
    fprintf(outfile, "%c", toupper(aoutname[i]));
  fprintf(outfile,"\n");

  inputbus = getvsym(&busnametable,"input_bus");
  if ((inputbus != NULL) && (inchannels > 0))
    {
      fprintf(outfile,"#define ASYS_INDRIVER_");
      j = 0;
      for (i=0; i< (int)strlen(ainname); i++)
	if (ainname[i] == '.')
	  {
	    j = i + 1;
	    break;
	  }
      for (i=j; i< (int)strlen(ainname); i++)
	fprintf(outfile, "%c", toupper(ainname[i]));
      fprintf(outfile,"\n");
    }

  /****************************/
  /* include actual libraries */
  /****************************/

  if (((ainflow == PASSIVE_FLOW) && (aoutflow == PASSIVE_FLOW)) || 
      (ain == aout) || (ainflow == ACTIVE_FLOW))
    {
      makeoutputaudiodriver(inputbus);
      makeinputaudiodriver(inputbus);
    }
  else
    {
      /* aout is active flow, define passive input driver first */ 

      makeinputaudiodriver(inputbus);
      makeoutputaudiodriver(inputbus);
    }

  fprintf(outfile,"\n");

  /*********************************************/
  /* include default ksync/ksyncinit if needed */
  /*********************************************/

  if ((makeaoutsync(aout) == 0) && (makeainsync(ain) == 0))
    {
      if (timeoptions == RENDER)
	{
	  fprintf(outfile,"\nfloat ksync() { return 0.0F; }\n\n");
	  fprintf(outfile,"\nvoid ksyncinit(void) { }\n\n");
	}
      if (timeoptions == TIMESYNC)
	maketsync();
      if (timeoptions == PLAYBACK)
	makepsync();
    }

  fprintf(outfile,"\n");

}

/****************************************************************/
/*             loads in appropriate control library             */
/****************************************************************/

void controllibrary(void)

{

  if ((cin == 0) && (session == 0))
    return;

  fprintf(outfile,"#define CSYS_DONE 0\n");
  fprintf(outfile,"#define CSYS_ERROR 1\n\n");
  fprintf(outfile,"#define CSYS_NONE 0\n");

  if (cmidi || session)
    fprintf(outfile,"#define CSYS_MIDIEVENTS 1\n");

  if (csasl)
    fprintf(outfile,"#define CSYS_SASLEVENTS 2\n");

  fprintf(outfile,"#define CSYS_EVENTS 3\n\n");

  if (session)
    {
      fprintf(outfile,"#define NSYS_DONE 0\n");
      fprintf(outfile,"#define NSYS_NONE 0\n");
      fprintf(outfile,"#define NSYS_MIDIEVENTS 1\n");
    }

  if (cmidi || session)
    {
      fprintf(outfile,"#define CSYS_MIDI_NUMCHAN  16\n\n");

      fprintf(outfile,"#define CSYS_MIDI_SPECIAL  0x70\n");
      fprintf(outfile,"#define CSYS_MIDI_NOOP     0x70\n");
      fprintf(outfile,"#define CSYS_MIDI_NEWTEMPO 0x71\n");
      fprintf(outfile,"#define CSYS_MIDI_ENDTIME  0x72\n");
      fprintf(outfile,"#define CSYS_MIDI_POWERUP  0x73\n");
      fprintf(outfile,"#define CSYS_MIDI_NOTEOFF  0x80\n");
      fprintf(outfile,"#define CSYS_MIDI_NOTEON   0x90\n");
      fprintf(outfile,"#define CSYS_MIDI_PTOUCH   0xA0\n");
      fprintf(outfile,"#define CSYS_MIDI_CC       0xB0\n");
      fprintf(outfile,"#define CSYS_MIDI_PROGRAM  0xC0\n");
      fprintf(outfile,"#define CSYS_MIDI_CTOUCH   0xD0\n");
      fprintf(outfile,"#define CSYS_MIDI_WHEEL    0xE0\n");
      fprintf(outfile,"#define CSYS_MIDI_SYSTEM   0xF0\n\n");
      
      fprintf(outfile,"#define CSYS_MIDI_CC_BANKSELECT_MSB  0x00\n"); 
      fprintf(outfile,"#define CSYS_MIDI_CC_MODWHEEL_MSB    0x01\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_BREATHCNTRL_MSB 0x02\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_FOOTCNTRL_MSB   0x04\n"); 
      fprintf(outfile,"#define CSYS_MIDI_CC_PORTAMENTO_MSB  0x05\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_DATAENTRY_MSB   0x06\n"); 
      fprintf(outfile,"#define CSYS_MIDI_CC_CHANVOLUME_MSB  0x07\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_BALANCE_MSB     0x08\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_PAN_MSB         0x0A\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_EXPRESSION_MSB  0x0B\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT1_MSB     0x0C\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT2_MSB     0x0D\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN1_MSB        0x10\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN2_MSB        0x11\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN3_MSB        0x12\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN4_MSB        0x13\n"); 
      fprintf(outfile,"#define CSYS_MIDI_CC_BANKSELECT_LSB  0x20\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_MODWHEEL_LSB    0x21\n");       
      fprintf(outfile,"#define CSYS_MIDI_CC_BREATHCNTRL_LSB 0x22\n");       
      fprintf(outfile,"#define CSYS_MIDI_CC_FOOTCNTRL_LSB   0x24\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_PORTAMENTO_LSB  0x25\n");       
      fprintf(outfile,"#define CSYS_MIDI_CC_DATAENTRY_LSB   0x26\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_CHANVOLUME_LSB  0x27\n");       
      fprintf(outfile,"#define CSYS_MIDI_CC_BALANCE_LSB     0x28\n");    
      fprintf(outfile,"#define CSYS_MIDI_CC_PAN_LSB         0x2A\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_EXPRESSION_LSB  0x2B\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT1_LSB     0x2C\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT2_LSB     0x2D\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN1_LSB        0x30\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN2_LSB        0x31\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN3_LSB        0x32\n");  
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN4_LSB        0x33\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SUSTAIN         0x40\n");        
      fprintf(outfile,"#define CSYS_MIDI_CC_PORTAMENTO      0x41\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_SUSTENUTO       0x42\n"); 
      fprintf(outfile,"#define CSYS_MIDI_CC_SOFTPEDAL       0x43\n"); 
      fprintf(outfile,"#define CSYS_MIDI_CC_LEGATO          0x44\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_HOLD2           0x45\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL1   0x46\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL2   0x47\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL3   0x48\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL4   0x49\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL5   0x4A\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL6   0x4B\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL7   0x4C\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL8   0x4D\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL9   0x4E\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_SOUNDCONTROL10  0x4F\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN5            0x50\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN6            0x51\n");   
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN7            0x52\n");  
      fprintf(outfile,"#define CSYS_MIDI_CC_GEN8            0x53\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_PORTAMENTOSRC   0x54\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT1DEPTH    0x5B\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT2DEPTH    0x5C\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT3DEPTH    0x5D\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT4DEPTH    0x5E\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_EFFECT5DEPTH    0x5F\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_DATAENTRYPLUS   0x60\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_DATAENTRYMINUS  0x61\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_ALLSOUNDOFF     0x78\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_RESETALLCONTROL 0x79\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_LOCALCONTROL    0x7A\n");
      fprintf(outfile,"#define CSYS_MIDI_CC_ALLNOTESOFF     0x7B\n");
      fprintf(outfile,"\n");
    }

  if (csasl)
    {
      fprintf(outfile,"#define CSYS_SASL_INSTR    0x00\n");
      fprintf(outfile,"#define CSYS_SASL_CONTROL  0x01\n");
      fprintf(outfile,"#define CSYS_SASL_TABLE    0x02\n");
      fprintf(outfile,"#define CSYS_SASL_ENDTIME  0x04\n");
      fprintf(outfile,"#define CSYS_SASL_TEMPO    0x05\n");
      fprintf(outfile,"#define CSYS_SASL_NOOP     0x06\n");
      fprintf(outfile,"\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_SAMPLE       0x6F\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_DATA         0x70\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_RANDOM       0x71\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_STEP         0x72\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_LINESEG      0x73\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_EXPSEG       0x74\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_CUBICSEG     0x75\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_POLYNOMIAL   0x76\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_SPLINE       0x77\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_WINDOW       0x78\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_HARM         0x79\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_HARM_PHASE   0x7A\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_PERIODIC     0x7B\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_BUZZ         0x7C\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_CONCAT       0x7D\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_EMPTY        0x7E\n");
      fprintf(outfile,"#define CSYS_SASL_TGEN_DESTROY      0x7F\n");
      fprintf(outfile,"\n");

    }

  if (cin)
    makecontroldriver(cin);

  if (session)
    makenetworkdriver();

  fprintf(outfile,"\n");

}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                                                              */
/* Utility functions, used by second level preamble() calls     */
/*                                                              */
/*______________________________________________________________*/


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* The first set of functions prints out the variables and      */
/* table symbols used by opcodes and its decendents. Used       */
/* for global, instr, and effects opcode symbol tables.         */
/*______________________________________________________________*/


/****************************************************************/
/*       set ibus values for oparraycalls -- recursive          */
/****************************************************************/

void opcodearrayidx(tnode * optr, int * j)

{
  int k = 1;

  while (optr != NULL)
    {
      switch (optr->ttype) {
      case S_OPARRAYDECL:
	opcodearrayidx(optr->sptr->defnode->optr,&k);
	k = 1;
	break;
      case S_OPCALL:
	opcodearrayidx(optr->sptr->defnode->optr,&k);
	k = 1;
	optr->arrayidx = *j;
	*j = *j + 1;
	break;
      case S_OPARRAYCALL:
	opcodearrayidx(optr->sptr->defnode->optr,&k);
	k = 1;
	optr->arrayidx = *j;
	*j = *j + 1;
      default:
	break;
      }
      optr = optr->next;
    }

}


/****************************************************************/
/*       prints one set of variables for oparray defines        */
/****************************************************************/

void oparrayslice(tnode * optr, char * prefix,
		  int * i, int opidx)

{  
  int k = 1;
  int m;
  sigsym * sptr;
  char newprefix[STRSIZE];

  /* no longer saves memory space for inlined core opcodes */
  /* and for mirrored variables, for consistent alignment  */
  /* note that optr->optr may be NULL (const-optimized)    */

  sptr = optr->sptr->defnode->sptr;
  while (sptr != NULL)
    {
      if (sptr->vartype != TMAPTYPE)
	{
	  fprintf(outfile,"#define %s_%soparray%i_%s %i\n",
		  prefix, optr->val, opidx, sptr->val,*i);
	  if (truewidth(sptr->width) == 0)  
	    {  
	      printf("Error: Zero-width arrays ");
	      printf("not permitted.\n");
	      showerrorplace(sptr->defnode->down->linenum,
			     sptr->defnode->down->filename);
	    }
	  *i = *i + truewidth(sptr->width);
	}	
      sptr = sptr->next;
    }

  if (!coreopcodename(optr))
    fprintf(outfile, "#define %s_%soparray%i__sym_opstate %i\n", 
	    prefix, optr->val, opidx, (*i)++);

  for (m = 0; m < optr->sptr->maxifstate; m++)
    fprintf(outfile, "#define %s_%soparray%i__sym_if%i %i\n", 
	    prefix, optr->val, opidx, m, (*i)++);

  sprintf(newprefix,"%s_%soparray%i",prefix,optr->val,opidx);
  opcodedefines(optr->sptr->defnode->optr,newprefix,i,&k);

}

extern int opstate_needed(tnode * optr);

/****************************************************************/
/*   print of #ifdef for opcodes -- recursive                   */
/****************************************************************/

void opcodedefines(tnode * optr, char * prefix,
		   int * i, int * j)

{
  sigsym * sptr;
  char newprefix[STRSIZE];
  int k = 1;
  int opnum, opsize, m;

  while (optr != NULL)
    {
      switch (optr->ttype) {
      case S_OPARRAYDECL:
	curroparraydepth++;
	opsize = 0;		    
	fprintf(outfile,"#define %s_%sopbase %i\n",prefix,optr->val, *i);
	oparrayslice(optr, prefix, &opsize, 0);
	fprintf(outfile,"#define %s_%sopsize %i\n",prefix,optr->val,opsize);
	for (opnum = 0; opnum < truewidth(optr->opwidth);opnum++)
	  oparrayslice(optr, prefix, i, opnum + 1);
	curroparraydepth--;
	break;
      case S_OPCALL:

	/* no longer saves memory space for inlined core opcodes */
	/* and for mirrored variables, for consistent alignment, */
	/* once we hit the first oparray in the stack.           */
	/* note that optr->optr may be NULL (const-optimized)    */

	if (curroparraydepth || 
	    (optr->optr && (!coreopcodecaninline(optr->optr->down))))
	  {
	    sptr = optr->sptr->defnode->sptr;
	    while (sptr != NULL)
	      {
		if ((sptr->vartype != TMAPTYPE) && 
		    (curroparraydepth || (sptr->tref->mirror == REQUIRED)))
		  {
		    fprintf(outfile,"#define %s_%s%i_%s %i\n",prefix,
			    optr->val,*j,sptr->val,*i);
		    if (truewidth(sptr->width) == 0)  
		      {   
			/* will be caught before this */
			printf("Error: Zero-width arrays not permitted.\n");
			showerrorplace(sptr->defnode->down->linenum,
				       sptr->defnode->down->filename);
		      }
		    *i = *i + truewidth(sptr->width);
		  }
		sptr = sptr->next;
	      }
	    
	    if (opstate_needed(optr))
	      fprintf(outfile,"#define %s_%s%i__sym_opstate %i\n", prefix,
		      optr->val,*j, (*i)++);

	    for (m = 0; m < (curroparraydepth ? optr->sptr->maxifstate 
			     : optr->sptr->cref->ifstate); m++)
	      fprintf(outfile,"#define %s_%s%i__sym_if%i %i\n", prefix,
		      optr->val,*j, m, (*i)++);
	  }

	if (curroparraydepth || (optr->optr && 
				 (!polyopcallexcept(optr->optr->down))
				 && (optr->sptr->cref->callrate > optr->rate)))
	  fprintf(outfile,"#define %s_%s%i__sym_ocstate %i\n",
		  prefix, optr->val, *j, (*i)++);

	fprintf(outfile,"#define %s_%s%i_return %i\n",prefix,optr->val,*j,*i);
	*i = *i + truewidth(optr->width);
	sprintf(newprefix,"%s_%s%i",prefix,optr->val,*j);
	opcodedefines(optr->sptr->defnode->optr,newprefix,i,&k);
	k = 1;
	optr->arrayidx = *j;
	*j = *j + 1;
	break;
      case S_OPARRAYCALL:

	/*  no calls to opcodedefine() --> no curroparraydepth{++,--} */

	if (curroparraydepth || (optr->optr && 
				 (!polyopcallexcept(optr->optr->down))
				 && (optr->sptr->cref->callrate > optr->rate)))
	  fprintf(outfile,"#define %s_%s%i__sym_ocstate %i\n",
		  prefix, optr->val, *j, (*i)++);

	fprintf(outfile,"#define %s_%s%i_return %i\n",prefix,optr->val,*j,*i);
	*i = *i + truewidth(optr->width);
	opcodearrayidx(optr->sptr->defnode->optr,&k);
	k = 1;
	optr->arrayidx = *j;
	*j = *j + 1;
      default:
	break;
      }
      optr = optr->next;
    }

}


/****************************************************************/
/*       prints one set of tables for oparray defines           */
/****************************************************************/

void oparraytableslice(tnode * optr, char * prefix,
		  int * i, int opidx)

{    
  int k = 1;
  sigsym * sptr;
  char newprefix[STRSIZE];

  sptr = optr->sptr->defnode->sptr;
  while (sptr != NULL)
    {
      if ((sptr->vartype == TABLETYPE)&&(sptr->kind != K_PFIELD))
	{
	  fprintf(outfile,"#define TBL_%s_%soparray%i_%s %i\n",
		  prefix,optr->val,(opidx),sptr->val,*i);
	  *i = *i + 1;
	}
      sptr = sptr->next;
    }
  sprintf(newprefix,"%s_%soparray%i",prefix,optr->val,opidx);
  opcodetabledefines(optr->sptr->defnode->optr,newprefix,i,&k);

}

/****************************************************************/
/*   print of #ifdef for opcode tables -- recursive             */
/****************************************************************/

void opcodetabledefines(tnode * optr,char * prefix,
			int * i, int * j)

{
  sigsym * sptr;
  char newprefix[STRSIZE];
  int k = 1;
  int opnum, opsize;

  while (optr != NULL)
    {
      switch (optr->ttype) {
      case S_OPARRAYDECL:

	opsize = 0;		    
	fprintf(outfile,"#define TBL_%s_%sopbase %i\n",prefix,optr->val, *i);
	oparraytableslice(optr, prefix, &opsize, 0);
	fprintf(outfile,"#define TBL_%s_%sopsize %i\n", prefix, optr->val,
		opsize);

	for (opnum = 0; opnum < truewidth(optr->opwidth);opnum++)
	  oparraytableslice(optr, prefix, i, opnum + 1);
	break;
      case S_OPCALL:
	sptr = optr->sptr->defnode->sptr;
	while (sptr != NULL)
	  {
	    if ((sptr->vartype == TABLETYPE)&&(sptr->kind != K_PFIELD))
	      {
		fprintf(outfile,"#define TBL_%s_%s%i_%s %i\n",
			prefix,optr->val, *j,sptr->val,*i);
		*i = *i + 1;
	      }
	    sptr = sptr->next;
	  }
	sprintf(newprefix,"%s_%s%i",prefix,optr->val,*j);
	opcodetabledefines(optr->sptr->defnode->optr,newprefix,i,&k);
	k = 1;
	*j = *j + 1;
	break;
      case S_OPARRAYCALL:
	*j = *j + 1;
      default:
	break;
      }
      optr = optr->next;
    }

}

/****************************************************************/
/*     determines if the opstate state variable is needed       */
/****************************************************************/

int opstate_needed(tnode * optr)

{
  if (!coreopcodename(optr))
    {
      if (curroparraydepth)
	return 1;

      switch (optr->rate) {
      case KRATETYPE:
	return (optr->sptr->cref->ilines > 0);
	break;
      case ARATETYPE:
	return (optr->sptr->cref->klines || optr->sptr->cref->ilines);
	break;
      }
    }

  return 0;
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* These functions are utilities for printing out instr symbols */
/*______________________________________________________________*/

/****************************************************************/
/*            prints control driver instr symbols               */
/****************************************************************/

void instrdriversymbols(sigsym * sptr, sigsym * iptr, int i)

{
  sigsym * bptr;
  sigsym * b2ptr;

 if ((sptr->kind == K_IMPORT) ||(sptr->kind == K_IMPORTEXPORT))
   {
     if (bitfile && bitsymin)
       {
	 bptr = getvsym(&bitsymin,sptr->val);
	 b2ptr = getvsym(&bitsymin,iptr->val);
	 
	 /* null for synthetic global tables */
	 
	 if (bptr && b2ptr)
	   fprintf(outfile,
		   "#define CSYS_SASL_IMPORT_%s_%s %i\n",
		   b2ptr->defnode->val, bptr->defnode->val,
		   i);
       }
     else
       if ((sptr->vartype!=TABLETYPE)||(sptr->val[0]!='_'))
	 fprintf(outfile,
		 "#define CSYS_SASL_IMPORT_%s_%s %i\n",
		 iptr->val,sptr->val,i);
   }
}

/****************************************************************/
/*  calculates width for symbols -- unclear if differences from */
/*  truewidth() are important, so keeping it for now.           */
/****************************************************************/

int symbolwidth(int width)

{

  if ((width == INCHANNELSWIDTH) || (width == CHANNELSWIDTH))
    return currinputwidth;
  if (width == OUTCHANNELSWIDTH)
    return currinstrwidth;
  return width;

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* This function family prints the external function headers    */
/*           needed later in the sa.c file.                     */
/*______________________________________________________________*/


/****************************************************************/
/*                 print opcode declarations                    */
/****************************************************************/


void printopexterns(tnode * tptr)

{

  char * stack;
  char newfix[STRSIZE];
  char newname[STRSIZE];
  int opnum;

  while (tptr != NULL)
    {
      if (tptr->optr && ((tptr->ttype == S_OPARRAYCALL) ||   
	  ((tptr->ttype == S_OPCALL) &&
	   (!coreopcodecaninline(tptr->optr->down)))))
	{
	  sprintf(newfix,"%s_%s%i",currinstancename, tptr->val, 
		  tptr->arrayidx);
	  sprintf(newname,"%s__sym_%s%i",currinstancename, tptr->val, 
		  tptr->arrayidx);
	  if (currinstance == NULL)
	    {
	      fprintf(outfile,
		      "float %s(struct ninstr_types *);\n", newname);
	      if (tptr->special)
		fprintf(outfile,
			"float %s_spec(struct ninstr_types *);\n",
			newname);
	    }
	  else
	    {
	      fprintf(outfile,"float %s(void);\n", newname);
	      if (tptr->special)
		fprintf(outfile,"float %s_spec(void);\n", newname);
	    }

	  /* do constant oparray index syntax check */

	  if ((tptr->ttype == S_OPARRAYCALL) && 
	      (tptr->optr->down->next->next->vol == CONSTANT))
	    {
	      opnum = (int) make_long(tptr->optr->down->next->next->down);
		  
	      if ((opnum < 0) || (opnum >= truewidth(tptr->ibus->opwidth)))
		{
		  printf("Error: Oparray call %s constant index %i out of range.\n\n",
			 tptr->val, opnum);
		  showerrorplace(tptr->optr->down->linenum, 
				 tptr->optr->down->filename);
		}
	    }

	  /* identical naming scheme for oparrays and opcalls */

	  stack = currinstancename;
	  currinstancename = newfix;
	  printopexterns(tptr->sptr->defnode->optr);
	  currinstancename = stack;

	}
      tptr = tptr->next;
    }

}


/****************************************************************/
/*            prints pass-function externs                     */
/****************************************************************/

void printpassexterns(void)

{

  sigsym * sptr;
  tnode * tptr;

  sptr = instrnametable;
  while (sptr != NULL)
    {
      if ((sptr->score)||(sptr->ascore)||(sptr->dyn)||(sptr->midi)
	  ||(sptr->amidi)
	  ||((cmidi || session) && sptr->miditag) || csasl)
	{
	  fprintf(outfile,"void %s_ipass(struct ninstr_types *);\n",sptr->val);
	  fprintf(outfile,"void %s_kpass(struct ninstr_types *);\n",sptr->val);
	  if (sptr->cref->alines)
	    fprintf(outfile,"void %s_apass(struct ninstr_types *);\n",
		    sptr->val);
	}
      sptr = sptr->next;
    }
  fprintf(outfile,"\n");

  tptr = instances;
  while (tptr != NULL)
    {
      fprintf(outfile,"void %s_ipass(void);\n",tptr->val);
      fprintf(outfile,"void %s_kpass(void);\n",tptr->val);
      if (tptr->sptr->cref->alines)
	fprintf(outfile,"void %s_apass(void);\n",tptr->val);
      tptr = tptr->next;
    }
  fprintf(outfile,"\n");
}


/****************************************************************/
/*        print finput/finGroup  declarations                   */
/****************************************************************/

void printfinputexterns(void)

{

  tnode * tptr = instances;
  
  while (tptr != NULL)
    {
      fprintf(outfile,"float finput%i(float);\n",
	      tptr->arrayidx);
      fprintf(outfile,"float finGroup%i(float);\n",
	      tptr->arrayidx);
      tptr = tptr->next;
    }
  fprintf(outfile,"\n");
}


extern void printsaolopcodetables(tnode * optr, char * prefix, int type);

/****************************************************************/
/*     prints arrays externs for SAOL constant wavetables       */
/*           also called in writeorc.c to print table data      */
/****************************************************************/

void printsaoltables(int type)

{
  tnode * tptr;
  sigsym * sptr;
  sigsym * iptr;
  char name[STRSIZE];

  sptr = globalsymtable;

  while (sptr != NULL)
    {
      if ((sptr->vartype == TABLETYPE) && (sptr->kind == K_NORMAL) &&
	  (sptr->defnode->vol == CONSTANT) && (sptr->defnode->usesinput))
	{
	  if (type == S_DATA)
	    {	  
	      sprintf(name, "table_global_%s",sptr->val);
	      printtablestring(sptr, name);
	    }
	  else
	    fprintf(outfile,"extern %s table_global_%s[];\n", 
		    hexstrings ? "char" : "float", sptr->val);
	}
      sptr = sptr->next;
    }

  printsaolopcodetables(globalopcodecalls, "global", type); 

  iptr = instrnametable;
  while (iptr != NULL)
    {
      if ((iptr->score)||(iptr->ascore)||(iptr->dyn)||(iptr->startup)
	  ||(iptr->midi)||(iptr->amidi)
	  ||((cmidi || session) && iptr->miditag)|| csasl)
	{
	  sptr = iptr->defnode->sptr;
	  while (sptr != NULL)
	    {
	      if ((sptr->vartype == TABLETYPE) &&
		  (sptr->kind == K_NORMAL) &&
		  (sptr->defnode->vol == CONSTANT) &&
		  (sptr->defnode->usesinput))
		{
		  if (type == S_DATA)
		    {	  
		      sprintf(name, "table_%s_%s", iptr->val, sptr->val);
		      printtablestring(sptr, name);
		    }
		  else
		    fprintf(outfile,"extern %s table_%s_%s[];\n",
			    hexstrings ? "char" : "float", 
			    iptr->val, sptr->val);
		}
	      sptr = sptr->next;
	    }
	  printsaolopcodetables(iptr->defnode->optr, iptr->val, type);
	}
      iptr = iptr->next;
    }

  tptr = instances;
  while (tptr != NULL)
    {
      sptr = tptr->sptr->defnode->sptr;
      while (sptr != NULL)
	{
	  if (sptr->vartype == TABLETYPE)
	    {
	      if ((sptr->vartype == TABLETYPE) &&
		  (sptr->kind == K_NORMAL) &&
		  (sptr->defnode->vol == CONSTANT) &&
		  (sptr->defnode->usesinput))
		{
		  if (type == S_DATA)
		    {	  
		      sprintf(name, "table_%s_%s", tptr->val, sptr->val);
		      printtablestring(sptr, name);
		    }
		  else
		    fprintf(outfile,"extern %s table_%s_%s[];\n",
			    hexstrings ? "char" : "float", 
			    tptr->val, sptr->val);
		}
	    }
	  sptr = sptr->next;
	}

      printsaolopcodetables(tptr->sptr->defnode->optr, tptr->val, type);
      tptr = tptr->next;
    }

  fprintf(outfile,"\n");
}


/****************************************************************/
/*   print array externs and data for opcall constant tables    */
/****************************************************************/

void printsaolopcodetables(tnode * optr, char * prefix, int type)


{
  sigsym * sptr;
  char newprefix[STRSIZE];
  char name[STRSIZE];

  while (optr != NULL)
    {
      sprintf(newprefix,"%s_%s%i", prefix, optr->val, optr->arrayidx);
      if ((optr->ttype == S_OPCALL) || (optr->ttype == S_OPARRAYCALL))
	{
	  sptr = optr->sptr->defnode->sptr;
	  while (sptr != NULL)
	    {
	      if ((sptr->vartype == TABLETYPE) &&
		  (sptr->kind == K_NORMAL) &&
		  (sptr->defnode->vol == CONSTANT) &&
		  (sptr->defnode->usesinput))
		{
		  if (type == S_DATA)
		    {	  
		      sprintf(name, "table_%s_%s", newprefix, sptr->val);
		      printtablestring(sptr, name);
		    }
		  else
		    fprintf(outfile,"extern %s table_%s_%s[];\n",
			    hexstrings ? "char" : "float", 
			    newprefix, sptr->val);
		}
	      sptr = sptr->next;
	    }
	  printsaolopcodetables(optr->sptr->defnode->optr, newprefix, type);
	}
      optr = optr->next;
    }
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* This function family handles the reflection interface,       */
/*     that prints out SAOL program structure for the           */
/*                  control driver to view.                     */
/*       printcdriverpreamble() is main routine, at end         */  
/*______________________________________________________________*/


/****************************************************************/
/*   print datastructures for MIDI/SASL control preambles       */
/****************************************************************/

void printcdriverstructs(void)

{
  fprintf(outfile,"#define MAXCNOTES %i\n\n",MAXDNOTES);
  fprintf(outfile,"#define CSYS_INSTRNUM %i\n\n", numinstrnames);

  fprintf(outfile,"#define CSYS_IRATE 0\n");
  fprintf(outfile,"#define CSYS_KRATE 1\n");
  fprintf(outfile,"#define CSYS_ARATE 2\n");
  fprintf(outfile,"#define CSYS_TABLE 3\n\n");

  fprintf(outfile,"#define CSYS_NORMAL 0\n");
  fprintf(outfile,"#define CSYS_IMPORT 1\n");
  fprintf(outfile,"#define CSYS_EXPORT 2\n");
  fprintf(outfile,"#define CSYS_IMPORTEXPORT 3\n");
  fprintf(outfile,"#define CSYS_PFIELD 4\n");
  fprintf(outfile,"#define CSYS_INTERNAL 5\n\n");

  fprintf(outfile,"#define CSYS_STATUS_EFFECTS    1 \n");
  fprintf(outfile,"#define CSYS_STATUS_SCORE      2 \n");
  fprintf(outfile,"#define CSYS_STATUS_MIDI       4 \n");
  fprintf(outfile,"#define CSYS_STATUS_DYNAMIC    8 \n");
  fprintf(outfile,"#define CSYS_STATUS_STARTUP   16 \n\n");

  fprintf(outfile,"typedef struct csys_varstruct {\n");
  fprintf(outfile,"  int index;\n");
  fprintf(outfile,"  char * name;\n");
  fprintf(outfile,"  int token;\n");
  fprintf(outfile,"  int type;\n");
  fprintf(outfile,"  int tag;\n");
  fprintf(outfile,"  int width;\n");
  fprintf(outfile,"} csys_varstruct;\n\n");

  fprintf(outfile,"typedef struct csys_instrstruct {\n");
  fprintf(outfile,"  int index;\n");
  fprintf(outfile,"  char * name;\n");
  fprintf(outfile,"  int token;\n");
  fprintf(outfile,"  int numvars;\n");
  fprintf(outfile,"  csys_varstruct * vars;\n");
  fprintf(outfile,"  int outwidth;\n");
  fprintf(outfile,"  int status;\n");
  fprintf(outfile,"} csys_instrstruct;\n\n");

  fprintf(outfile,"typedef struct csys_labelstruct {\n");
  fprintf(outfile,"  int index;\n");
  fprintf(outfile,"  char * name;\n");
  fprintf(outfile,"  int token;\n");
  fprintf(outfile,"  int iflag[CSYS_INSTRNUM];\n");
  fprintf(outfile,"} csys_labelstruct;\n\n");

  fprintf(outfile,"typedef struct csys_presetstruct {\n");
  fprintf(outfile,"  int index;\n");
  fprintf(outfile,"  int preset;\n");
  fprintf(outfile,"} csys_presetstruct;\n\n");

  fprintf(outfile,"typedef struct csys_samplestruct {\n");
  fprintf(outfile,"  int index;\n");
  fprintf(outfile,"  int token;\n");
  fprintf(outfile,"  char * name;\n");
  fprintf(outfile,"  char * fname;\n");
  fprintf(outfile,"} csys_samplestruct;\n\n");

  fprintf(outfile,"typedef struct csys_routestruct {\n");
  fprintf(outfile,"  int bus;\n");     
  fprintf(outfile,"  int ninstr;\n");  
  fprintf(outfile,"  int * instr;\n"); 
  fprintf(outfile,"} csys_routestruct;\n\n");

  fprintf(outfile,"typedef struct csys_sendstruct {\n");
  fprintf(outfile,"  int instr;\n");  
  fprintf(outfile,"  int nbus;\n");   
  fprintf(outfile,"  int * bus;\n");  
  fprintf(outfile,"} csys_sendstruct;\n\n");

  fprintf(outfile,"typedef struct csys_busstruct {\n");  
  fprintf(outfile,"  int index;\n");     
  fprintf(outfile,"  char * name;\n");   
  fprintf(outfile,"  int width;\n");     
  fprintf(outfile,"  int oflag;\n");    
  fprintf(outfile,"} csys_busstruct;\n\n");  

  fprintf(outfile,"typedef struct csys_targetstruct {\n");    
  fprintf(outfile,"char * name;\n");    
  fprintf(outfile,"int token;\n");    
  fprintf(outfile,"int numinstr;\n");    
  fprintf(outfile,"int * instrindex;\n");    
  fprintf(outfile,"int * varindex;\n");    
  fprintf(outfile,"} csys_targetstruct;\n\n");  

}

/****************************************************************/
/*   print functions for MIDI/SASL control preambles       */
/****************************************************************/

void printcdriverfunctions(void)

{
  fprintf(outfile,"\n");  
  fprintf(outfile,"void csys_terminate(char * message)\n");  
  fprintf(outfile,"{\n");
  fprintf(outfile,"   epr(0,NULL,\"%s control driver\", message);\n",cinname);
  fprintf(outfile,"}\n\n");
}

/****************************************************************/
/*   print global variables for MIDI/SASL control preambles     */
/****************************************************************/

void printcdriverglobalvars(void)

{
  int numglobal = 0;
  int i, gcount, tbl;
  sigsym * sptr, * bptr;

  sptr = globalsymtable;
  while (sptr)
    {
      if ((sptr->vartype == SCALARTYPE) ||
	  (sptr->vartype == VECTORTYPE) ||
	  ((sptr->vartype == TABLETYPE) && (sptr->val[0] != '_')))
	numglobal++;
      sptr = sptr->next;
    }

  i =  MIDIFRAMELEN*totmidichan;

  fprintf(outfile,"#define CSYS_GLOBALNUM %i\n\n",numglobal);
  if (!numglobal)
    {
      fprintf(outfile,"csys_varstruct csys_global[1];\n\n");
    }
  else
    {
      fprintf(outfile,"csys_varstruct csys_global[CSYS_GLOBALNUM] = {\n");
      sptr = globalsymtable;
      gcount = numglobal;
      tbl = 0;
      while (sptr)
	{      
	  if ((sptr->vartype == SCALARTYPE) ||
	      (sptr->vartype == VECTORTYPE) ||
	      ((sptr->vartype == TABLETYPE) && (sptr->val[0] != '_')))
	    {
	      /* int index */

	      if (sptr->vartype == TABLETYPE)
		fprintf(outfile,"%i, ",tbl);  
	      else
		fprintf(outfile,"%i, ",i);  

	      /* char * name */
	      
	      if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
		fprintf(outfile,"\"%s\", ",bptr->defnode->val);
	      else
		fprintf(outfile,"\"%s\", ",sptr->val);

	      /* int token */

	      fprintf(outfile,"%s, ", bitfile ? &(sptr->val[1]) : "-1");

	      /* int type and int tag*/

	      switch (sptr->vartype) {
	      case SCALARTYPE: 
	      case VECTORTYPE: 
		if (sptr->rate == IRATETYPE)
		  fprintf(outfile,"CSYS_IRATE, CSYS_NORMAL, ");
		else
		  fprintf(outfile,"CSYS_KRATE, CSYS_NORMAL, ");
		break;
	      case TABLETYPE:
		fprintf(outfile,"CSYS_TABLE, CSYS_NORMAL, ");
		break;
	      }
	      
	      fprintf(outfile,"%i",sptr->width);  /* int width */
	      if (--gcount)
		fprintf(outfile,", \n");
	      else
		fprintf(outfile," };\n\n");
	    }
	  i = i + sptr->width;
	  if (sptr->vartype == TABLETYPE)
	    tbl++;
	  sptr = sptr->next;
	}
    }
}




/****************************************************************/
/*   computes position of labelled control variable in stack    */
/****************************************************************/

int targetvarposition(sigsym * sptr, sigsym * vptr)

{
  int ret = 0;

  while (sptr)
    {
      if (sptr == vptr)
	return ret;
      else
	if (sptr->vartype != TMAPTYPE)
	  ret += sptr->width;
      sptr = sptr->next;
    }

  /* should never execute */

  return 0;
}


/****************************************************************/
/*   print target variables for labelled SASL control statement */
/****************************************************************/

void printcdrivertarget(void)

{
  char * name;
  int numtarget, numinstr;
  int i, tcount;
  int * instrindex, * varindex;
  sigsym * sptr, * bptr, * iptr;

  sptr = targetsymtable;
  numtarget = 0;
  vmcheck(instrindex = (int *) calloc(numinstrnames, sizeof(int)));
  vmcheck(varindex = (int *) calloc(numinstrnames, sizeof(int)));
  while (sptr)
    {
      if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
	name = bptr->defnode->val;
      else
	name = sptr->val;

      iptr = instrnametable;
      numinstr = 0;
      while (iptr)
	{
	  if ((bptr = getvsym(&(iptr->defnode->sptr),sptr->val)) &&
	      (bptr->kind == K_IMPORT))
	    {
	      instrindex[numinstr] = iptr->vol;
	      varindex[numinstr] = targetvarposition(iptr->defnode->sptr,
						     bptr);
	      numinstr++;
	    }
	  iptr = iptr->next;
	}
      fprintf(outfile,"#define CSYS_TARGET_INSTR_%s %i\n\n",
	      name, numinstr);
      if (!numinstr)
	{
	  fprintf(outfile, "int csys_targetinstr_%s[1];\n", name);
	  fprintf(outfile, "int csys_targetvar_%s[1];\n\n",  name);
	}
      else
	{      
	  fprintf(outfile,
		  "int csys_targetinstr_%s[CSYS_TARGET_INSTR_%s] = {\n %i",
		  name, name, instrindex[0]);
	  for (i=1; i< numinstr; i++)
	    fprintf(outfile, ", %i",instrindex[i]);
	  fprintf(outfile, "};\n\n");
	  fprintf(outfile,
		  "int csys_targetvar_%s[CSYS_TARGET_INSTR_%s] = {\n %i",
		  name, name, varindex[0]);
	  for (i=1; i< numinstr; i++)
	    fprintf(outfile, ", %i",varindex[i]);
	  fprintf(outfile, "};\n\n");
	}
      numtarget++;
      sptr = sptr->next;
    }

  fprintf(outfile,"#define CSYS_TARGETNUM %i\n\n",numtarget);
  if (!numtarget)
    {
      fprintf(outfile,"csys_targetstruct csys_target[1];\n\n");
    }
  else
    {
      fprintf(outfile,"csys_targetstruct csys_target[CSYS_TARGETNUM] = {\n");
      sptr = targetsymtable;
      tcount = 0;
      while (sptr)
	{      

	  /* char * name */
	      
	  if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
	    name = bptr->defnode->val;
	  else
	    name = sptr->val;

	  fprintf(outfile,"\"%s\", ",name);

	  /* int token */

	  fprintf(outfile,"%s, ", bitfile ? &(sptr->val[1]) : "-1");

	  /* int numinstr */

	  fprintf(outfile,"CSYS_TARGET_INSTR_%s, ", name);

	  /* int numinstr */

	  fprintf(outfile,"&(csys_targetinstr_%s[0]), ", name);

	  /* int numinstr */

	  fprintf(outfile,"&(csys_targetvar_%s[0]) ", name);
	  if ((++tcount) == numtarget)
	     fprintf(outfile,"};\n\n");
	  else
	    fprintf(outfile,",\n");
	  sptr = sptr->next;
	}
    }

  free(instrindex);
  free(varindex);

}


/****************************************************************/
/*   print labels for MIDI/SASL control preambles               */
/****************************************************************/

void printcdriverlabels(void)

{
  sigsym * sptr, * bptr, * iptr;
  tnode * tptr;
  int found, totlabels;

  totlabels = allsasl->numlabels + abssasl->numlabels;

  fprintf(outfile,"#define CSYS_NOLABEL 0\n\n");
  fprintf(outfile,"#define CSYS_LABELNUM %i\n\n", totlabels);
  
  if (!(totlabels)) 
    fprintf(outfile,"csys_labelstruct csys_labels[1];\n\n");
  else
    {
      sptr = allsasl->labeltable;
      fprintf(outfile,"csys_labelstruct csys_labels[CSYS_LABELNUM] = {\n");
      while (sptr)
	{
	  /* int index */

	  fprintf(outfile, "%i, ",sptr->special+1); 

	  /* char * name */

	  if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
	    fprintf(outfile,"\"%s\", ",bptr->defnode->val);
	  else
	    fprintf(outfile,"\"%s\", ",sptr->val);

	  /* int token */

	  fprintf(outfile,"%s, ", bitfile ? &(sptr->val[1]) : "-1");

	  iptr = instrnametable;
	  while (iptr)
	    {
	      found = 0;
	      tptr = sptr->defnode;
	      while (tptr)
		{
		  if (tptr->down && (tptr->down->sptr == iptr))
		    {
		      found = 1;
		      break;
		    }
		  tptr = tptr->next;
		}
	      fprintf(outfile, "%i ",found);
	      iptr = iptr->next;
	      if (iptr)
		fprintf(outfile, ", ");
	    }
	  sptr = sptr->next;
	  if (sptr)
	    fprintf(outfile,",\n");
	  else
	    fprintf(outfile,"};\n\n");
	}
    }

}

/****************************************************************/
/*   print preset numbers used in instrs                        */
/****************************************************************/

void printcdriverpresets(void)

{
  sigsym * sptr;
  int i;

  i = 0;
  sptr = instrpresets;
  while (sptr)
    {
      i++;
      sptr = sptr->next;
    }

  fprintf(outfile,"#define CSYS_PRESETNUM %i\n\n", i);
  
  if (!i) 
    fprintf(outfile,"csys_presetstruct csys_presets[1];\n\n");
  else
    {
      sptr = instrpresets;
      fprintf(outfile,"csys_presetstruct csys_presets[CSYS_PRESETNUM] = {\n");
      while (sptr)
	{
	  /* int index */

	  fprintf(outfile, "%i, ",sptr->defnode->sptr->vol); 

	  /* int preset */

	  fprintf(outfile, "%i ",sptr->width); 

	  sptr = sptr->next;
	  if (sptr)
	    fprintf(outfile,",\n");
	  else
	    fprintf(outfile,"};\n\n");
	}
    }

}

/****************************************************************/
/*   print sample files used in configuration block             */
/****************************************************************/

void printcdriversamples(void)

{
  sigsym * sptr, * bptr;
  char * name;
  int i;

  i = 0;
  sptr = bitsamplein;
  while (sptr)
    {
      i++;
      sptr = sptr->next;
    }

  fprintf(outfile,"#define CSYS_SAMPLENUM %i\n\n", i);
  
  if (!i) 
    fprintf(outfile,"csys_samplestruct csys_samples[1];\n\n");
  else
    {
      i = 0;
      sptr = bitsamplein;
      fprintf(outfile,"csys_samplestruct csys_samples[CSYS_SAMPLENUM] = {\n");
      while (sptr)
	{
	  /* int index */

	  fprintf(outfile, "%i, ",i++); 

	  /* int token */

	  fprintf(outfile, "%s, ",&(sptr->val[1])); 

	  /* char * name */

	  if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
	    name = bptr->defnode->val;
	  else
	    name = sptr->val;
  
	  fprintf(outfile, "\"%s\", ", name); 

	  /* char * fname */

	  fprintf(outfile, "\"%s.wav\" ", sptr->val); 

	  sptr = sptr->next;
	  if (sptr)
	    fprintf(outfile,",\n");
	  else
	    fprintf(outfile,"};\n\n");
	}
    }

}

/****************************************************************/
/*   prints csys_varstruct for control driver instrs            */
/****************************************************************/

void printcdrivervars(sigsym * instr)

{

  sigsym * sptr, * bptr;
  int numvars, i, comma;
  char * name, * varname;

  sptr = instr->defnode->sptr;
  numvars = 0;
  while (sptr)
    {
      /* don't include tablemaps or internal variables */
      
      if ((sptr->vartype != TMAPTYPE) && (sptr->val[0] != '_')) 
	numvars++;
      sptr = sptr->next;
    }
  
  if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,instr->val)))
    name = bptr->defnode->val;
  else
    name = instr->val;
  
  fprintf(outfile,"#define CSYS_%s_VARNUM %i\n\n",name,numvars);
  if (!numvars) 
    fprintf(outfile,"csys_varstruct csys_%s_vars[1];\n\n",name);
  else
    {
      fprintf(outfile,"csys_varstruct csys_%s_vars[CSYS_%s_VARNUM] = {\n",
	      name, name);
      comma = i = 0;
      currinputwidth = 1;              /* makes dummy location */
      currinstrwidth = instr->width;
      sptr = instr->defnode->sptr;

      while (sptr)
	{
	  if ((sptr->vartype != TMAPTYPE) && (sptr->val[0] != '_'))
	    {	
	      if (comma)
		fprintf(outfile,",\n");
	      comma = 1;

	      if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
		varname = bptr->defnode->val;
	      else
		varname = sptr->val;

	      /* int index */

	      fprintf(outfile,"%i, ", i);

	      /* char * name */

	      fprintf(outfile,"\"%s\", ", varname);

	      /* int token */

	      fprintf(outfile,"%s, ", bitfile ? &(sptr->val[1]) : "-1");

	      /* int type */

	      switch (sptr->vartype) {
	      case SCALARTYPE: 
	      case VECTORTYPE: 
		switch (sptr->rate) {
		case IRATETYPE:
		  fprintf(outfile,"CSYS_IRATE, ");
		  break;
		case KRATETYPE:
		  fprintf(outfile,"CSYS_KRATE, ");
		  break;
		case ARATETYPE:
		  fprintf(outfile,"CSYS_ARATE, ");
		  break;
		}
		break;
	      case TABLETYPE:
		fprintf(outfile,"CSYS_TABLE, ");
		break;
	      }

	      /* int tag */

	      switch (sptr->kind) {
	      case K_NORMAL:
		fprintf(outfile,"CSYS_NORMAL, ");
		break;
	      case K_PFIELD:
		fprintf(outfile,"CSYS_PFIELD, ");
		break;
	      case K_IMPORT:
		fprintf(outfile,"CSYS_IMPORT, ");
		break;
	      case K_EXPORT:
		fprintf(outfile,"CSYS_EXPORT, ");
		break;
	      case K_IMPORTEXPORT:
		fprintf(outfile,"CSYS_IMPORTEXPORT, ");
		break;
	      default:
		fprintf(outfile,"CSYS_INTERNAL, ");
		break;
	      }

	      /* int width */
	      fprintf(outfile,"%i ",truewidth(sptr->width));  

	    }

	  if (sptr->vartype != TMAPTYPE)
	    i += truewidth(sptr->width);

	  sptr = sptr->next;
	}
      fprintf(outfile," };\n\n");
    }
}

/****************************************************************/
/*   prints csys_inststruct for control driver instrs           */
/****************************************************************/

void printcdriverinstr(sigsym * sptr)

{
  char * name;
  sigsym * bptr;
  int status;

  if (sptr->vol)
    fprintf(outfile,",\n");

  if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
    name = bptr->defnode->val;
  else
    name = sptr->val;
  
  /* int index, char * name */

  fprintf(outfile,"%i, \"%s\",",sptr->vol, name);  
    
  /* int token */

  fprintf(outfile,"%s, ", bitfile && !(sptr->startup) ?
	  &(sptr->val[1]) : "-1");

  /* numvars, * vars */
  fprintf(outfile,"CSYS_%s_VARNUM, &(csys_%s_vars[0]),",
	  name, name);

  /* output width */

  fprintf(outfile,"%i, ",sptr->width);

  /* status word */

  status = 0;
  if (sptr->effects)
    status |=  STATWORD_EFFECTS;

  if (sptr->score || sptr->ascore)
    status |=  STATWORD_SCORE;

  if (sptr->midi || sptr->amidi)
    status |=  STATWORD_MIDI;

  if (sptr->dyn)
    status |=  STATWORD_DYNAMIC;

  if (sptr->startup)
    status |=  STATWORD_STARTUP;


  fprintf(outfile,"%i ",status);

}

/****************************************************************/
/*           print buses in SAOL program                        */
/****************************************************************/

void printcdriverbus(void)

{
  char * name;
  sigsym * sptr, * bptr;
  tnode * tptr;
  int i;

  i = 0;
  sptr = busnametable;
  while (sptr)
    {
      i++;
      sptr = sptr->next;
    }

  fprintf(outfile,"#define CSYS_BUSNUM %i\n\n", i);
  
  if (!i) 
    fprintf(outfile,"csys_busstruct csys_bus[1];\n\n");
  else
    {
      sptr = busnametable;
      fprintf(outfile,"csys_busstruct csys_bus[CSYS_BUSNUM] = {\n");
      while (sptr)
	{
	  /* int index */

	  fprintf(outfile, "%i, ",sptr->vol);
 
	  if (bitfile && bitsymin && (bptr = getvsym(&bitsymin,sptr->val)))
	    name = bptr->defnode->val;
	  else
	    name = sptr->val;
  
	  /* char * name */

	  fprintf(outfile, "\"%s\",", name); 

	  /* int width */

	  fprintf(outfile,"%i, ",sptr->width);

	  /* int oflag */

	  i = 0;
	  tptr = outbustable;
	  while (tptr)
	    {
	      if (tptr->sptr == sptr)
		i++;
	      tptr = tptr->next;
	    }
	  fprintf(outfile,"%i ", i);
	  
	  sptr = sptr->next;
	  if (sptr)
	    fprintf(outfile,",\n");
	  else
	    fprintf(outfile,"};\n\n");
	}
    }

}

/****************************************************************/
/*           print route statements in SAOL program             */
/****************************************************************/

void printcdriverroute(void)

{
  tnode * tptr, * iptr;
  int i, j;

  i = 0;
  tptr = groot;
  while (tptr)
    {
      if (tptr->ttype == S_ROUTEDEF)
	{
	  j = 0;
	  iptr = tptr->down->next->next->next->next->down;
	  while (iptr)
	    {
	      if (iptr->ttype == S_IDENT)
		j++;
	      iptr = iptr->next;
	    }
	  fprintf(outfile,"#define CSYS_ROUTE_INUM_%i %i\n\n",
		  i, j);
	  fprintf(outfile,"int csys_route%i_instr[CSYS_ROUTE_INUM_%i] = {\n",
		  i,i);
	  iptr = tptr->down->next->next->next->next->down;
	  while (iptr)
	    {
	      if (iptr->ttype == S_IDENT)
		{
		  fprintf(outfile," %i",tptr->sptr->vol);
		  if (iptr->next)
		    fprintf(outfile,",");
		  else
		    fprintf(outfile,"};\n\n");
		}
	      iptr = iptr->next;
	    }
	  i++;
	}
      tptr = tptr->next;
    }

  fprintf(outfile,"#define CSYS_ROUTENUM %i\n\n", i);
  
  if (!i) 
    fprintf(outfile,"csys_routestruct csys_route[1];\n\n");
  else
    {
      tptr = groot;
      i = 0;
      fprintf(outfile,"csys_routestruct csys_route[CSYS_ROUTENUM] = {\n");
      while (tptr)
	{
	  if (tptr->ttype == S_ROUTEDEF)
	    {
	      if (i)
		fprintf(outfile, "\n, ");

	      /* int bus */

	      fprintf(outfile, "%i, ",tptr->sptr->vol);
 
	      /* int ninstr */

	      fprintf(outfile, "CSYS_ROUTE_INUM_%i, ",i);
	      
	      /* int * instr */

	      fprintf(outfile, "&(csys_route%i_instr[0]) ",i);

	      i++;
	    }
	  tptr = tptr->next;
	}
      fprintf(outfile,"};\n\n");
    }

}

/****************************************************************/
/*           print send statements in SAOL program             */
/****************************************************************/

void printcdriversend(void)

{
  tnode * tptr, * iptr;
  int i, j;

  i = 0;
  tptr = groot;
  while (tptr)
    {
      if (tptr->ttype == S_SENDDEF)
	{
	  j = 0;
	  iptr = tptr->down->next->next->next->next->next->next->down;
	  while (iptr)
	    {
	      if (iptr->ttype == S_IDENT)
		j++;
	      iptr = iptr->next;
	    }
	  fprintf(outfile,"#define CSYS_SEND_BNUM_%i %i\n\n",
		  i, j);
	  fprintf(outfile,"int csys_send%i_bus[CSYS_SEND_BNUM_%i] = {\n",
		  i,i);
	  iptr = tptr->down->next->next->next->next->next->next->down;
	  while (iptr)
	    {
	      if (iptr->ttype == S_IDENT)
		{
		  fprintf(outfile," %i",tptr->sptr->vol);
		  if (iptr->next)
		    fprintf(outfile,",");
		  else
		    fprintf(outfile,"};\n\n");
		}
	      iptr = iptr->next;
	    }
	  i++;
	}
      tptr = tptr->next;
    }

  fprintf(outfile,"#define CSYS_SENDNUM %i\n\n", i);
  
  if (!i) 
    fprintf(outfile,"csys_sendstruct csys_send[1];\n\n");
  else
    {
      tptr = groot;
      i = 0;
      fprintf(outfile,"csys_sendstruct csys_send[CSYS_SENDNUM] = {\n");
      while (tptr)
	{
	  if (tptr->ttype == S_SENDDEF)
	    {
	      if (i)
		fprintf(outfile, "\n, ");

	      /* int instr */

	      fprintf(outfile, "%i, ",tptr->sptr->vol);
 
	      /* int nbus */

	      fprintf(outfile, "CSYS_SEND_BNUM_%i, ",i);
	      
	      /* int * instr */

	      fprintf(outfile, "&(csys_send%i_bus[0]) ",i);

	      i++;
	    }
	  tptr = tptr->next;
	}
      fprintf(outfile,"};\n\n");
    }

}

/****************************************************************/
/*                print audioBIFS variables       */
/****************************************************************/

void printcdriveraudiobifs(void)

{
  fprintf(outfile," float position[3];\n");
  fprintf(outfile," float direction[3];\n");
  fprintf(outfile," float listenerPosition[3];\n");
  fprintf(outfile," float listenerDirection[3];\n");
  fprintf(outfile," float minFront;\n");
  fprintf(outfile," float maxFront;\n");
  fprintf(outfile," float minBack;\n");
  fprintf(outfile," float maxBack;\n");
  fprintf(outfile," float params[128];\n\n");

}


/****************************************************************/
/*   print preamble for MIDI and SASL control drivers       */
/****************************************************************/

void printcdriverpreamble(void)

{
  sigsym * sptr;

  if ((cin == 0) && (session == 0))
    {
      /* SAOL programs may try to write params, even if no     */
      /* control driver is specified, so we create it & leave. */

      if (has.s_params)
	fprintf(outfile,"float params[128];\n\n");
      return;
    }

  printcdriverstructs();      /* structs for control driver API */
  printcdriveraudiobifs();    /* audioBIFs standard names       */
  printcdriverfunctions();    /* functions for API              */
  printcdriverglobalvars();   /* fill global variable info      */
  printcdrivertarget();       /* fill target info               */
  printcdriverlabels();       /* fill SASL label info           */
  printcdriverpresets();      /* fill MIDI preset info          */
  printcdriversamples();      /* fill sample file info          */
  printcdriverbus();          /* fill bus info                  */
  printcdriverroute();        /* fill route info                */
  printcdriversend();         /* fill send info                 */

  /* during this pass, make vars array and param array */

  sptr = instrnametable;
  while (sptr != NULL)
    {
      printcdrivervars(sptr);
      sptr = sptr->next;
    }      

  /* during this pass, fill in complete data structure */

  fprintf(outfile,"csys_instrstruct csys_instr[CSYS_INSTRNUM] = {\n");

  sptr = instrnametable;
  while (sptr != NULL)
    {
      printcdriverinstr(sptr);
      sptr = sptr->next;
    }     

  fprintf(outfile,"};\n\n");

}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*  This function family prints out data lists for the          */
/*     various kinds of instrument data sources.                */
/*______________________________________________________________*/


/****************************************************************/
/*            prints initialized lists for instrs               */
/****************************************************************/

void printinstrlists(int * totlines, int * usesdyn, int * instrnum)

{
  sigsym * sptr;
  sigsym * bptr;
  char * val;

  sptr = instrnametable;
  while (sptr != NULL)
    {
      if (sptr->score)
	{
	  initscoreinstr(RELTSTAMP, sptr);
	  (*totlines) += sptr->score;
	}
      if (sptr->ascore)
	{
	  initscoreinstr(ABSTSTAMP, sptr);
	  (*totlines) += sptr->ascore;
	}
      if (sptr->midi)
	{
	  initmidiinstr(RELTSTAMP, sptr, totlines);
	}
      if (sptr->amidi)
	{
	  initmidiinstr(ABSTSTAMP, sptr, totlines);
	}
      if (sptr->dyn)
	{
	  *usesdyn = 1;
	}
      if (sptr->effects)
	{
	  fprintf(outfile,"struct instr_line e_%s[%i];\n\n",sptr->val,
		  sptr->effects);
	  (*totlines) += sptr->effects;
	}
      if (sptr->startup)
	{
	  fprintf(outfile,"struct instr_line u_startup[1];\n");
	  (*totlines)++;
	}
      if ((cmidi || session) && sptr->miditag)
	{
	  val = dupunderscore(sptr->val);
	  fprintf(outfile,"struct instr_line cm_%s[MAXCNOTES];\n",
		  sptr->val);
	  fprintf(outfile,"struct instr_line * cm_%sfirst = &cm_%s[1];\n",
		  val,sptr->val);
	  fprintf(outfile,"struct instr_line * cm_%slast = &cm_%s[0];\n",
		  val,sptr->val);
	  fprintf(outfile,
		  "struct instr_line * cm_%send = &cm_%s[MAXCNOTES-1];\n",
		  val,sptr->val);
	  fprintf(outfile,"struct instr_line * cm_%snext = NULL;\n\n",
		  val);
	  free(val);
	}

      if (csasl)
	{
	  if (bitfile && bitsymin)
	    {
	      bptr = getvsym(&bitsymin,sptr->val);
	      if (bptr)
		fprintf(outfile,"#define CSYS_SASL_INSTR_%s %i\n\n",
			bptr->defnode->val, (*instrnum)++);
	      else
		internalerror("writepre.c", "bitsymin");
	    }
	  else
	    {
	      fprintf(outfile,"#define CSYS_SASL_INSTR_%s %i\n\n",
		      sptr->val,(*instrnum)++);
	    }
	  val = dupunderscore(sptr->val);
	  fprintf(outfile,"struct instr_line cs_%s[MAXCNOTES];\n",
		  sptr->val);
	  fprintf(outfile,"struct instr_line * cs_%sfirst = &cs_%s[1];\n",
		  val,sptr->val);
	  fprintf(outfile,"struct instr_line * cs_%slast = &cs_%s[0];\n",
		  val,sptr->val);
	  fprintf(outfile,
		  "struct instr_line * cs_%send = &cs_%s[MAXCNOTES-1];\n",
		  val,sptr->val);
	  fprintf(outfile,"struct instr_line * cs_%snext = NULL;\n\n",
		  val);
	  free(val);
	}
      sptr = sptr->next;
    }

}


/****************************************************************/
/*                 prints cm declarations for presets           */
/****************************************************************/

void printpresetinit(char * name)

{
  int i;
  int j = 0;
  sigsym * sptr = instrpresets;
  char * val;

  fprintf(outfile,"struct instr_line **cmp_%s[CSYS_MAXPRESETS] = {\n",name);
  for (i = 0; i <= maxmidipreset;i++)
    {
      if ((sptr == NULL) || (i != sptr->width))
	fprintf(outfile,"NULL");
      else
	{
	  val = dupunderscore(sptr->defnode->sptr->val);
	  fprintf(outfile,"&cm_%s%s",val,name);
	  free(val);
	  sptr = sptr->next;
	}
      if (i != maxmidipreset)
	fprintf(outfile,",");
      if (!((++j)%5))
	fprintf(outfile,"\n");
    }
  fprintf(outfile,"};\n\n");

}

/****************************************************************/
/*                 prints cm declarations for presets           */
/****************************************************************/

void printextchaninit(char * name)

{
  int i;
  int j = 0;
  sigsym * sptr = NULL;
  char * val;

  fprintf(outfile,"struct instr_line **cme_%s[CSYS_MAXEXTCHAN] = {\n",name);
  for (i = 0; i < cmaxchan; i++)
    {
      if ((i & 15) == 0)
	sptr = instrpresets;
      if (sptr == NULL)
	fprintf(outfile,"NULL");
      else
	{
	  val = dupunderscore(sptr->defnode->sptr->val);
	  fprintf(outfile,"&cm_%s%s",val,name);
	  free(val);
	  sptr = sptr->next ? sptr->next : instrpresets;
	}
      if (i != cmaxchan-1)
	fprintf(outfile,",");
      if (!((++j)%5))
	fprintf(outfile,"\n");
    }
  fprintf(outfile,"};\n\n");

}

/****************************************************************/
/*            prints cm declaration for preset number           */
/****************************************************************/

void printextchanpreset(void)

{
  int i;
  int j = 0;
  sigsym * sptr = instrpresets;

  fprintf(outfile,"int cme_preset[CSYS_MAXEXTCHAN] = {\n");
  for (i = 0; i < cmaxchan; i++)
    {
      if ((sptr == NULL) || (i != sptr->width))
	fprintf(outfile,"CSYS_MAXPRESETS");
      else
	{
	  fprintf(outfile,"%i",i);
	  sptr = sptr->next;
	}
      if (i != cmaxchan-1)
	fprintf(outfile,",");
      if (!((++j)%5))
	fprintf(outfile,"\n");
    }
  fprintf(outfile,"};\n\n");

}

/****************************************************************/
/*                 prints cs declarations for presets           */
/****************************************************************/

void printsaslinit(char * name)

{
  int j = 0;
  sigsym * sptr = instrnametable;
  char * val;

  fprintf(outfile,"struct instr_line **cs_%s[CSYS_MAXSASLINSTR] = {\n",name);
  while (sptr != NULL)
    {
      val = dupunderscore(sptr->val);
      fprintf(outfile,"&cs_%s%s",val,name);
      free(val);
      sptr = sptr->next;
      if (sptr != NULL)
	fprintf(outfile,",");
      if (!((++j)%5))
	fprintf(outfile,"\n");
    }
  fprintf(outfile,"};\n\n");

  if (strcmp(name,"first") && strcmp(name,"last"))
    return;

  if (abssasl->instrroot)
    {
      j = 0;
      sptr = instrnametable;
      fprintf(outfile,
	      "struct instr_line **csa_%s[CSYS_MAXSASLINSTR] = {\n",name);
      while (sptr != NULL)
	{
	  if (sptr->ascore)
	    {
	      val = dupunderscore(sptr->val);
	      fprintf(outfile,"&sa_%s%s",val,name);
	      free(val);
	    }
	  else
	    fprintf(outfile,"NULL");

	  sptr = sptr->next;
	  if (sptr != NULL)
	    fprintf(outfile,",");
	  if (!((++j)%5))
	    fprintf(outfile,"\n");
	}
      fprintf(outfile,"};\n\n");
    }

  if (allsasl->instrroot)
    {
      j = 0;
      sptr = instrnametable;
      fprintf(outfile,
	      "struct instr_line **css_%s[CSYS_MAXSASLINSTR] = {\n",name);
      while (sptr != NULL)
	{
	  if (sptr->score)
	    {
	      val = dupunderscore(sptr->val);
	      fprintf(outfile,"&s_%s%s",val,name);
	      free(val);
	    }
	  else
	    fprintf(outfile,"NULL");
	  sptr = sptr->next;
	  if (sptr != NULL)
	    fprintf(outfile,",");
	  if (!((++j)%5))
	    fprintf(outfile,"\n");
	}
      fprintf(outfile,"};\n\n");
    }

}



/****************************************************************/
/*   print declarations for MIDI and SASL control drivers       */
/****************************************************************/

void printcdriverlists(int * instrnum)

{

  if (cmidi || session)
    {
      fprintf(outfile,"#define CSYS_CCPOS %i\n", MIDICTRLPOS);
      fprintf(outfile,"#define CSYS_TOUCHPOS %i\n", MIDITOUCHPOS);
      fprintf(outfile,"#define CSYS_CHTOUCHPOS %i\n", MIDICHTOUCHPOS);
      fprintf(outfile,"#define CSYS_BENDPOS %i\n", MIDIBENDPOS);
      fprintf(outfile,"#define CSYS_EXTPOS %i\n", MIDIEXTPOS);
      fprintf(outfile,"#define CSYS_FRAMELEN %i\n", MIDIFRAMELEN);

      fprintf(outfile,"#define CSYS_MAXPRESETS %i\n\n",maxmidipreset+1);
      fprintf(outfile,"#define CSYS_NULLPROGRAM %i\n\n", null_program);
      printpresetinit("first");
      printpresetinit("last");
      printpresetinit("end");
      printpresetinit("next");

      fprintf(outfile,"#define CSYS_MAXEXTCHAN %i\n\n", cmaxchan);
      printextchaninit("first");
      printextchaninit("last");
      printextchaninit("end");
      printextchaninit("next");
      printextchanpreset();

      fprintf(outfile,"int csys_bank = 0;\n");
      fprintf(outfile,"int csys_banklsb = 0;\n");
      fprintf(outfile,"int csys_bankmsb = 0;\n\n");
    }

  if (csasl)
    {
      fprintf(outfile,"\n#define CSYS_MAXSASLINSTR %i\n",*instrnum);
      fprintf(outfile,"#define CSYS_SASL_NOINSTR %i\n",*instrnum);
      fprintf(outfile,"#define CSYS_SASL_MAXCONTROL %i\n",MAXDCONTROL);
      fprintf(outfile,"#define CSYS_SASL_MAXTABLES %i\n\n",MAXDTABLES);
      printsaslinit("first");
      printsaslinit("last");
      printsaslinit("end");
      printsaslinit("next");
      fprintf(outfile,"struct {\n");
      fprintf(outfile,"unsigned short id;\n");
      fprintf(outfile,"unsigned short label;\n");
      fprintf(outfile,"unsigned int fptr;\n");
      fprintf(outfile,"float fval;\n");
      fprintf(outfile,"} saslcontrol[CSYS_SASL_MAXCONTROL];\n\n");
      fprintf(outfile,"int maxsc = 0;\n\n");
      fprintf(outfile,"typedef struct table_cdtype {\n");
      fprintf(outfile,"unsigned short index; /* gtables index */\n");
      fprintf(outfile,"unsigned short tgen;  /* table generator */\n");
      fprintf(outfile,"unsigned int pnum;  /* number of params */\n");
      fprintf(outfile,"float * p;            /* param list */\n");
      fprintf(outfile,"} table_cdtype;\n");
      fprintf(outfile,"table_cdtype sasltable[CSYS_SASL_MAXTABLES];\n\n");
      fprintf(outfile,"int maxtb = 0;\n\n");
      maketgen();
    }
      
}


/****************************************************************/
/*   print declarations for dynamic instruments                 */
/****************************************************************/

void printdynamiclists(void)

{
  sigsym * sptr;
  char * val;

  sptr = instrnametable;
  while (sptr != NULL)
    {
      if (sptr->dyn)
	{
	  val = dupunderscore(sptr->val); 
	  fprintf(outfile,"struct instr_line d_%s[MAXLINES];\n"
		  ,sptr->val);
	  fprintf(outfile,"struct instr_line * d_%sfirst = &d_%s[1];\n"
		  ,val,sptr->val);
	  fprintf(outfile,"struct instr_line * d_%slast = &d_%s[0];\n"
		  ,val,sptr->val);
	  fprintf(outfile,
		  "struct instr_line * d_%send = &d_%s[MAXLINES-1];\n"
		  ,val,sptr->val);
	  fprintf(outfile,"struct instr_line * d_%snext = NULL;\n"
		  ,val);
	  free(val); 
	}
      sptr = sptr->next;
    }
}


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*   These functions print out audio input and output driver.   */
/*    The order in which these functions are flow-dependent.    */
/*______________________________________________________________*/

/****************************************************************/
/*               make the output audio driver                   */
/****************************************************************/

void makeoutputaudiodriver(sigsym * inputbus)

{ 
  int dosync;

  fprintf(outfile,"#define ASYS_HASOUTPUT\n"); 

  if ((ain == aout) && (inputbus != NULL) && (inchannels > 0))
    fprintf(outfile,"#define ASYS_HASINPUT\n");

  dosync = makeaoutsync(aout);

  if (dosync)
    fprintf(outfile,"#define ASYS_KSYNC\n");

  fprintf(outfile,"\n");

  makeaudiodriver(aout);

  fprintf(outfile,"#undef ASYS_HASOUTPUT\n");

  if (dosync)
    fprintf(outfile,"#undef ASYS_KSYNC\n");
}

/****************************************************************/
/*               make the output audio driver                   */
/****************************************************************/

void makeinputaudiodriver(sigsym * inputbus)

{  
  int dosync;

  if ((ain != aout) && (inputbus != NULL) && (inchannels > 0))
    {
      fprintf(outfile,"#define ASYS_HASINPUT\n");

      dosync = ((!makeaoutsync(aout)) && makeainsync(ain));
      
      if (dosync)
	fprintf(outfile,"#define ASYS_KSYNC\n");
      
      fprintf(outfile,"\n");
      makeaudiodriver(ain);
      fprintf(outfile,"\n");
      fprintf(outfile,"#undef ASYS_HASINPUT\n");

      if (dosync)
	fprintf(outfile,"#undef ASYS_KSYNC\n");
    }
}

